# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb ldap feed source"""

import ldap3
from logilab.common.configuration import merge_options

from yams import ValidationError
from cubicweb import AuthenticationError, Binary
from cubicweb import _
from cubicweb.server import utils
from cubicweb.server.sources import datafeed

# search scopes
LDAP_SCOPES = {
    "BASE": ldap3.BASE,
    "ONELEVEL": ldap3.LEVEL,
    "SUBTREE": ldap3.SUBTREE,
}

# map ldap protocol to their standard port
PROTO_PORT = {
    "ldap": 389,
    "ldaps": 636,
    "ldapi": None,
}


def replace_filter(s):
    s = s.replace("*", "\\2A")
    s = s.replace("(", "\\28")
    s = s.replace(")", "\\29")
    s = s.replace("\\", "\\5c")
    s = s.replace("\0", "\\00")
    return s


class LDAPFeedSource(datafeed.DataFeedSource):
    """LDAP feed source: unlike ldapuser source, this source is copy based and
    will import ldap content (beside passwords for authentication) into the
    system source.
    """

    support_entities = {"CWUser": False}
    use_cwuri_as_url = False

    options = (
        (
            "auth-mode",
            {
                "type": "choice",
                "default": "simple",
                "choices": ("simple", "digest_md5", "gssapi"),
                "help": "authentication mode used to authenticate user to the ldap.",
                "group": "ldap-source",
                "level": 3,
            },
        ),
        (
            "auth-realm",
            {
                "type": "string",
                "default": None,
                "help": "realm to use when using gssapi/kerberos authentication.",
                "group": "ldap-source",
                "level": 3,
            },
        ),
        (
            "data-cnx-dn",
            {
                "type": "string",
                "default": "",
                "help": "user dn to use to open data connection to the ldap (eg used \
to respond to rql queries). Leave empty for anonymous bind",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "data-cnx-password",
            {
                "type": "string",
                "default": "",
                "help": "password to use to open data connection to the ldap (eg used to respond to rql queries). Leave empty for anonymous bind.",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-base-dn",
            {
                "type": "string",
                "default": "",
                "help": "base DN to lookup for users; disable user importation mechanism if unset",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-scope",
            {
                "type": "choice",
                "default": "ONELEVEL",
                "choices": ("BASE", "ONELEVEL", "SUBTREE"),
                "help": (
                    'user search scope (valid values: "BASE", "ONELEVEL", "SUBTREE")'
                ),
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-classes",
            {
                "type": "csv",
                "default": ("top", "posixAccount"),
                "help": 'classes of user (with Active Directory, you want to say "user" here)',
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-filter",
            {
                "type": "string",
                "default": "",
                "help": (
                    "additional filters to be set in the ldap query to find valid users"
                ),
                "group": "ldap-source",
                "level": 2,
            },
        ),
        (
            "start-tls",
            {
                "type": "choice",
                "choices": ("true", "false"),
                "default": "false",
                "help": "Start tls on connection (before bind)",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-login-attr",
            {
                "type": "string",
                "default": "uid",
                "help": 'attribute used as login on authentication (with Active Directory, you want to use "sAMAccountName" here)',
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-default-group",
            {
                "type": "csv",
                "default": ("users",),
                "help": "name of a group in which ldap users will be by default. \
You can set multiple groups by separating them by a comma.",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "user-attrs-map",
            {
                "type": "named",
                "default": {"uid": "login"},
                "help": "map from ldap user attributes to cubicweb attributes (with Active Directory, you want to use sAMAccountName:login,mail:email,givenName:firstname,sn:surname)",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "group-base-dn",
            {
                "type": "string",
                "default": "",
                "help": "base DN to lookup for groups; disable group importation mechanism if unset",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "group-scope",
            {
                "type": "choice",
                "default": "ONELEVEL",
                "choices": ("BASE", "ONELEVEL", "SUBTREE"),
                "help": (
                    'group search scope (valid values: "BASE", "ONELEVEL", "SUBTREE")'
                ),
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "group-classes",
            {
                "type": "csv",
                "default": ("top", "posixGroup"),
                "help": "classes of group",
                "group": "ldap-source",
                "level": 1,
            },
        ),
        (
            "group-filter",
            {
                "type": "string",
                "default": "",
                "help": "additional filters to be set in the ldap query to find valid groups",
                "group": "ldap-source",
                "level": 2,
            },
        ),
        (
            "group-attrs-map",
            {
                "type": "named",
                "default": {"cn": "name", "memberUid": "member"},
                "help": "map from ldap group attributes to cubicweb attributes",
                "group": "ldap-source",
                "level": 1,
            },
        ),
    )

    options = merge_options(
        datafeed.DataFeedSource.options + options,
        optgroup="ldap-source",
    )

    _conn = None

    def check_urls(self, source_entity):
        urls = super(LDAPFeedSource, self).check_urls(source_entity)

        if len(urls) > 1:
            raise ValidationError(
                source_entity.eid, {"url": _("can only have one url")}
            )

        try:
            protocol, hostport = urls[0].split("://")
        except ValueError:
            raise ValidationError(source_entity.eid, {"url": _("badly formatted url")})
        if protocol not in PROTO_PORT:
            raise ValidationError(source_entity.eid, {"url": _("unsupported protocol")})

        return urls

    def init(self, source_entity):
        super(LDAPFeedSource, self).init(source_entity)
        typedconfig = self.config
        self.authmode = typedconfig["auth-mode"]
        self._authenticate = getattr(self, f"_auth_{self.authmode}")
        self.cnx_dn = typedconfig["data-cnx-dn"]
        self.cnx_pwd = typedconfig["data-cnx-password"]
        self.start_tls = typedconfig["start-tls"] == "true"
        self.user_base_dn = str(typedconfig["user-base-dn"])
        self.user_base_scope = LDAP_SCOPES[typedconfig["user-scope"]]
        self.user_login_attr = typedconfig["user-login-attr"]
        self.user_default_groups = typedconfig["user-default-group"]
        self.user_attrs = {"dn": "eid", "modifyTimestamp": "modification_date"}
        self.user_attrs.update(typedconfig["user-attrs-map"])
        self.user_rev_attrs = dict((v, k) for k, v in self.user_attrs.items())
        self.base_filters = [
            f"(objectclass={replace_filter(o)})" for o in typedconfig["user-classes"]
        ]
        if typedconfig["user-filter"]:
            self.base_filters.append(typedconfig["user-filter"])
        self.group_base_dn = str(typedconfig["group-base-dn"])
        self.group_base_scope = LDAP_SCOPES[typedconfig["group-scope"]]
        self.group_attrs = typedconfig["group-attrs-map"]
        self.group_attrs = {"dn": "eid", "modifyTimestamp": "modification_date"}
        self.group_attrs.update(typedconfig["group-attrs-map"])
        self.group_rev_attrs = dict((v, k) for k, v in self.group_attrs.items())
        self.group_base_filters = [
            f"(objectClass={replace_filter(o)})" for o in typedconfig["group-classes"]
        ]
        if typedconfig["group-filter"]:
            self.group_base_filters.append(typedconfig["group-filter"])
        self._conn = None

    def connection_info(self):
        assert len(self.urls) == 1, self.urls
        protocol, hostport = self.urls[0].split("://")
        if protocol != "ldapi" and ":" in hostport:
            host, port = hostport.rsplit(":", 1)
        else:
            host, port = hostport, PROTO_PORT[protocol]
        return protocol, host, port

    def authenticate(self, cnx, login, password=None, **kwargs):
        """return CWUser eid for the given login/password if this account is
        defined in this source, else raise `AuthenticationError`

        two queries are needed since passwords are stored crypted, so we have
        to fetch the salt first
        """
        self.info("ldap authenticate %s", login)
        if not password:
            # On Windows + ADAM this would have succeeded (!!!)
            # You get Authenticated as: 'NT AUTHORITY\ANONYMOUS LOGON'.
            # we really really don't want that
            raise AuthenticationError()
        searchfilter = [
            f"({replace_filter(self.user_login_attr)}={replace_filter(login)})"
        ]
        searchfilter.extend(self.base_filters)
        searchstr = f"(&{''.join(searchfilter)})"
        # first search the user
        try:
            user = self._search(
                cnx, self.user_base_dn, self.user_base_scope, searchstr
            )[0]
        except IndexError:
            # no such user
            raise AuthenticationError()
        # check password by establishing a (unused) connection
        try:
            self._connect(user, password)
        except ldap3.core.exceptions.LDAPException as ex:
            # Something went wrong, most likely bad credentials
            self.info("while trying to authenticate %s: %s", user, ex)
            raise AuthenticationError()
        except Exception:
            self.error("while trying to authenticate %s", user, exc_info=True)
            raise AuthenticationError()
        rset = cnx.execute(
            "Any X,SN WHERE X cwuri %(extid)s, X is CWUser, "
            "X cw_source S, S name SN",
            {"extid": user["dn"]},
        )
        if not rset or rset[0][1] != self.uri:
            # user is not known or has been moved away from this source
            raise AuthenticationError()
        return rset[0][0]

    def _connect(self, user=None, userpwd=None):
        protocol, host, port = self.connection_info()
        kwargs = {}
        if user:
            kwargs["user"] = user["dn"]
        elif self.cnx_dn:
            kwargs["user"] = self.cnx_dn
            if self.cnx_pwd:
                kwargs["password"] = self.cnx_pwd
        self.info(
            "connecting %s://%s:%s as %s",
            protocol,
            host,
            port,
            kwargs.get("user", "anonymous"),
        )
        server = ldap3.Server(host, port=int(port))
        conn = ldap3.Connection(
            server,
            client_strategy=ldap3.RESTARTABLE,
            auto_referrals=False,
            raise_exceptions=True,
            **kwargs,
        )
        if self.start_tls:
            conn.start_tls()

        # Now bind with the credentials given. Let exceptions propagate out.
        if user is None:
            # anonymous bind
            if not self.cnx_dn:
                if not conn.bind():
                    raise AuthenticationError(conn.result["message"])
            else:
                self._authenticate(conn, {"dn": self.cnx_dn}, self.cnx_pwd)
        else:
            # user specified, we want to check user/password, no need to return
            # the connection which will be thrown out
            if not self._authenticate(conn, user, userpwd):
                raise AuthenticationError()
        return conn

    def _auth_simple(self, conn, user, userpwd):
        conn.user = user["dn"]
        conn.password = userpwd
        return conn.bind()

    def _auth_digest_md5(self, conn, user, userpwd):
        conn.authentication = ldap3.AUTH_SASL
        conn.sasl_mechanism = "DIGEST-MD5"
        # realm, user, password, authz-id
        conn.sasl_credentials = (None, user["dn"], userpwd, None)
        return conn.bind()

    def _auth_gssapi(self, conn, user, userpwd):
        conn.authentication = ldap3.AUTH_SASL
        conn.sasl_mechanism = "GSSAPI"
        return conn.bind()

    def _search(self, cnx, base, scope, searchstr="(objectClass=*)", attrs=()):
        """make an ldap query"""
        self.debug(
            "ldap search %s %s %s %s %s", self.uri, base, scope, searchstr, list(attrs)
        )
        if self._conn is None:
            self._conn = self._connect()
        ldapcnx = self._conn
        if self.start_tls:
            ldapcnx.start_tls()
            self.info("ldap start_tls started for %s", self.uri)
        if not ldapcnx.search(
            base, searchstr, search_scope=scope, attributes=set(attrs) - {"dn"}
        ):
            return []
        result = []
        for rec in ldapcnx.response:
            if rec["type"] != "searchResEntry":
                continue
            items = rec["attributes"].items()
            itemdict = self._process_ldap_item(rec["dn"], items)
            result.append(itemdict)
        self.debug("ldap built results %s", len(result))
        return result

    def _process_ldap_item(self, dn, iterator):
        """Turn an ldap received item into a proper dict."""
        itemdict = {"dn": dn}
        for key, value in iterator:
            if self.user_attrs.get(key) == "upassword":  # XXx better password detection
                value = value[0]
                # we only support ldap_salted_sha1 for ldap sources, see: server/utils.py
                if not value.startswith(b"{SSHA}"):
                    value = utils.crypt_password(value)
                itemdict[key] = Binary(value)
            elif self.user_attrs.get(key) == "modification_date":
                itemdict[key] = value
            else:
                if hasattr(value, "__len__") and len(value) == 1:
                    itemdict[key] = value = value[0]
                else:
                    itemdict[key] = value
        # we expect memberUid to be a list of user ids, make sure of it
        member = self.group_rev_attrs["member"]
        if isinstance(itemdict.get(member), str):
            itemdict[member] = [itemdict[member]]
        return itemdict

    def _process_no_such_object(self, cnx, dn):
        """Some search return NO_SUCH_OBJECT error, handle this (usually because
        an object whose dn is no more existent in ldap as been encountered).

        Do nothing by default, let sub-classes handle that.
        """



try:
    uri, newdn = __args__
except ValueError:
    print(
        "USAGE: cubicweb-ctl shell <instance> ldap_change_base_dn.py -- <ldap source uri> <new dn>"
    )
    print()
    print("you should not have updated your sources file yet")

olddn = repo.source_by_uri(uri).config["user-base-dn"]

assert olddn != newdn

input("Ensure you've stopped the instance, type enter when done.")

for eid, olduserdn in rql(
    "Any X, XURI WHERE X cwuri XURI, X cw_source S, S name %(name)s", {"name": uri}
):
    newuserdn = olduserdn.replace(olddn, newdn)
    if newuserdn != olduserdn:
        print(olduserdn, "->", newuserdn)
        sql(f"UPDATE cw_cwuser SET cw_cwuri='{newuserdn}' WHERE cw_eid={eid}")

commit()

print("you can now update the sources file to the new dn and restart the instance")



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Exceptions shared by different cubicweb packages."""

from logilab.common.decorators import cachedproperty
from logilab.common.registry import RegistryException

from yams import ValidationError  # noqa: F401

# abstract exceptions #########################################################


class CubicWebException(Exception):
    """base class for cubicweb server exception"""

    msg = ""

    def __str__(self):
        if self.msg:
            if self.args:
                return self.msg % tuple(self.args)
            else:
                return self.msg
        else:
            return " ".join(str(arg) for arg in self.args)


class ConfigurationError(CubicWebException):
    """a misconfiguration error"""


class InternalError(CubicWebException):
    """base class for exceptions which should not occur"""


class SecurityError(CubicWebException):
    """base class for cubicweb server security exceptions"""


class RepositoryError(CubicWebException):
    """base class for repository exceptions"""


class SourceException(CubicWebException):
    """base class for source exceptions"""


class CubicWebRuntimeError(CubicWebException):
    """base class for runtime exceptions"""


# repository exceptions #######################################################


class ConnectionError(RepositoryError):
    """raised when a bad connection id is given or when an attempt to establish
    a connection failed
    """


class AuthenticationError(ConnectionError):
    """raised when an attempt to establish a connection failed due to wrong
    connection information (login / password or other authentication token)
    """


class BadConnectionId(ConnectionError):
    """raised when a bad connection id is given"""


class UnknownEid(RepositoryError):
    """the eid is not defined in the system tables"""

    msg = "No entity with eid %s in the repository"


class UniqueTogetherError(RepositoryError):
    """raised when a unique_together constraint caused an IntegrityError"""

    def __init__(self, session, **kwargs):
        self.session = session
        assert "rtypes" in kwargs or "cstrname" in kwargs
        self.kwargs = kwargs
        # fill cache while the session is open
        self.rtypes

    @cachedproperty
    def rtypes(self):
        if "rtypes" in self.kwargs:
            return self.kwargs["rtypes"]
        cstrname = str(self.kwargs["cstrname"])
        cstr = self.session.find("CWUniqueTogetherConstraint", name=cstrname).one()
        return sorted(rtype.name for rtype in cstr.relations)


class ViolatedConstraint(RepositoryError):
    def __init__(self, cnx, cstrname, query):
        self.cnx = cnx
        self.cstrname = cstrname
        message = (
            "constraint '%s' is being violated by the query '%s'. "
            "You can run the inverted constraint on the database to list the problematic rows."
        ) % (cstrname, query)
        super().__init__(message)


# security exceptions #########################################################


class Unauthorized(SecurityError):
    """raised when a user tries to perform an action without sufficient
    credentials
    """

    msg = "You are not allowed to perform this operation"
    msg1 = "You are not allowed to perform %s operation on %s"
    var = None

    def __str__(self):
        try:
            if self.args and len(self.args) == 2:
                return self.msg1 % self.args
            if self.args:
                return " ".join(self.args)
            return self.msg
        except Exception as ex:
            return str(ex)


class Forbidden(SecurityError):
    """raised when a user tries to perform a forbidden action"""


# source exceptions ###########################################################


class EidNotInSource(SourceException):
    """trying to access an object with a particular eid from a particular
    source has failed
    """

    msg = "No entity with eid %s in %s"


# registry exceptions #########################################################


class UnknownProperty(RegistryException):
    """property found in database but unknown in registry"""


# query exception #############################################################


class QueryError(CubicWebRuntimeError):
    """a query try to do something it shouldn't"""


class NotAnEntity(CubicWebRuntimeError):
    """raised when get_entity is called for a column which doesn't contain
    a non final entity
    """


class MultipleResultsError(CubicWebRuntimeError):
    """raised when ResultSet.one() is called on a resultset with multiple rows
    of multiple columns.
    """


class NoResultError(CubicWebRuntimeError):
    """raised when no result is found but at least one is expected."""


class UndoTransactionException(QueryError):
    """Raised when undoing a transaction could not be performed completely.

    Note that :
      1) the partial undo operation might be acceptable
         depending upon the final application

      2) the undo operation can also fail with a `ValidationError` in
         cases where the undoing breaks integrity constraints checked
         immediately.

      3) It might be that neither of those exception is raised but a
         subsequent `commit` might raise a `ValidationError` in cases
         where the undoing breaks integrity constraints checked at
         commit time.

    :type txuuix: int
    :param txuuid: Unique identifier of the partially undone transaction

    :type errors: list
    :param errors: List of errors occurred during undoing
    """

    msg = "The following error(s) occurred while undoing transaction #%d : %s"

    def __init__(self, txuuid, errors):
        super().__init__(txuuid, errors)
        self.txuuid = txuuid
        self.errors = errors


# tools exceptions ############################################################


class ExecutionError(Exception):
    """server execution control error (already started, not running...)"""


# pylint: disable=W0611
from logilab.common.clcommands import BadCommandUsage  # noqa: E402, F401



from yams.buildobjs import EntityType, RelationType, RelationDefinition, String

from cubicweb.schema import PUB_SYSTEM_ENTITY_PERMS, PUB_SYSTEM_REL_PERMS


class CWPermission(EntityType):
    """entity type that may be used to construct some advanced security
    configuration
    """

    __permissions__ = PUB_SYSTEM_ENTITY_PERMS

    name = String(
        required=True,
        indexed=True,
        internationalizable=True,
        maxsize=100,
        description=("name or identifier of the permission"),
    )
    label = String(
        required=True,
        internationalizable=True,
        maxsize=100,
        description=(
            "distinct label to distinguate between other "
            "permission entity of the same name"
        ),
    )


class granted_permission(RelationType):
    """explicitly granted permission on an entity"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    # XXX cardinality = '*1'


class require_permission(RelationType):
    __permissions__ = PUB_SYSTEM_REL_PERMS


class require_group(RelationDefinition):
    """groups to which the permission is granted"""

    __permissions__ = PUB_SYSTEM_REL_PERMS
    subject = "CWPermission"
    object = "CWGroup"


class has_group_permission(RelationDefinition):
    """short cut relation for 'U in_group G, P require_group G' for efficiency
    reason. This relation is set automatically, you should not set this.
    """

    __permissions__ = PUB_SYSTEM_REL_PERMS
    subject = "CWUser"
    object = "CWPermission"



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Test tools for cubicweb"""

import atexit
import errno
import getpass
import glob
import os
import pickle
import shlex
import shutil
import subprocess
import sys
import tempfile
from functools import partial
from hashlib import sha1  # pylint: disable=E0611
from os.path import join, exists, split, isdir

import filelock
from logilab.common.decorators import cached, clear_cache

from cubicweb import schema

# db auto-population configuration #############################################

SYSTEM_ENTITIES = (
    schema.SCHEMA_TYPES
    | schema.INTERNAL_TYPES
    | schema.WORKFLOW_TYPES
    | set(
        (
            "CWGroup",
            "CWUser",
        )
    )
)
SYSTEM_RELATIONS = (
    schema.META_RTYPES
    | schema.WORKFLOW_RTYPES
    | schema.WORKFLOW_DEF_RTYPES
    | schema.SYSTEM_RTYPES
    | schema.SCHEMA_TYPES
    | set(("primary_email",))  # deducted from other relations
)

# content validation configuration #############################################

# validators are used to validate (XML, DTD, whatever) view's content
# validators availables are :
#  'dtd' : validates XML + declared DTD
#  'xml' : guarantees XML is well formed
#  None : do not try to validate anything

# {'vid': validator}
VIEW_VALIDATORS = {}


# cubicweb test configuration ##################################################

BASE_URL = "https://testing.cubicweb/"

DEFAULT_SOURCES = {
    "system": {
        "adapter": "native",
        "db-encoding": "UTF-8",
        "db-user": "admin",
        "db-password": "gingkow",
        "db-name": "tmpdb",
        "db-driver": "sqlite",
        "db-host": None,
    },
    "admin": {
        "login": "admin",
        "password": "gingkow",
    },
}
DEFAULT_PSQL_SOURCES = DEFAULT_SOURCES.copy()
DEFAULT_PSQL_SOURCES["system"] = DEFAULT_SOURCES["system"].copy()
DEFAULT_PSQL_SOURCES["system"]["db-driver"] = "postgres"
DEFAULT_PSQL_SOURCES["system"]["db-user"] = getpass.getuser()
DEFAULT_PSQL_SOURCES["system"]["db-password"] = None
# insert a dumb value as db-host to avoid unexpected connection to local server
DEFAULT_PSQL_SOURCES["system"]["db-host"] = "REPLACEME"


def turn_repo_off(repo):
    """Idea: this is less costly than a full re-creation of the repo object.
    off:
    * session are closed,
    * cnxsets are closed
    * system source is shutdown
    """
    if not repo._needs_refresh:
        for cnxset in repo.cnxsets:
            cnxset.close(True)
        repo.system_source.shutdown()
        repo._needs_refresh = True
        repo._has_started = False


def turn_repo_on(repo):
    """Idea: this is less costly than a full re-creation of the repo object.
    on:
    * cnxsets are connected
    * cache are cleared
    """
    if repo._needs_refresh:
        for cnxset in repo.cnxsets:
            cnxset.reconnect()
        repo.clear_caches()
        repo._needs_refresh = False


# test database handling #######################################################

DEFAULT_EMPTY_DB_ID = "__default_empty_db__"


class TestDataBaseHandler:
    DRIVER = None

    db_cache = {}
    explored_glob = set()

    def __init__(self, config, init_config=None):
        self.config = config
        self.init_config = init_config
        self._repo = None
        # pure consistency check
        assert self.system_source["db-driver"] == self.DRIVER

        # some handlers want to store info here, avoid a warning
        from cubicweb.server.sources.native import NativeSQLSource

        NativeSQLSource.options += (
            ("global-db-name", {"type": "string", "help": "for internal use only"}),
        )

    def _ensure_test_backup_db_dir(self):
        """Return path of directory for database backup.

        The function create it if necessary"""
        backupdir = join(self.config.apphome, "database")
        try:
            os.makedirs(backupdir)
        except Exception:
            if not isdir(backupdir):
                raise
        # Add a .nobackup file in the database directory
        # so that backup utilities, like Borg, do not backup this directory
        open(join(backupdir, ".nobackup"), "w").close()
        return backupdir

    def config_path(self, db_id):
        """Path for config backup of a given database id"""
        return self.absolute_backup_file(db_id, "config")

    def absolute_backup_file(self, db_id, suffix):
        """Path for config backup of a given database id"""
        # in case db name is an absolute path, we don't want to replace anything
        # in parent directories
        directory, basename = split(self.dbname)
        dbname = basename.replace("-", "_")
        assert "." not in db_id
        filename = join(directory, f"{dbname}-{db_id}.{suffix}")
        return join(self._ensure_test_backup_db_dir(), filename)

    def db_cache_key(self, db_id, dbname=None):
        """Build a database cache key for a db_id with the current config

        This key is meant to be used in the cls.db_cache mapping"""
        if dbname is None:
            dbname = self.dbname
        dbname = os.path.basename(dbname)
        dbname = dbname.replace("-", "_")
        return (self.config.apphome, dbname, db_id)

    def backup_database(self, db_id):
        """Store the content of the current database as <db_id>

        The config used are also stored."""
        backup_data = self._backup_database(db_id)
        config_path = self.config_path(db_id)
        # XXX we dump a dict of the config
        # This is an experimental to help config dependant setup (like BFSS) to
        # be propertly restored
        pdir = os.path.dirname(config_path)
        with tempfile.NamedTemporaryFile(dir=pdir, delete=False) as conf_file:
            conf_file.write(pickle.dumps(dict(self.config)))
        os.rename(conf_file.name, config_path)
        self.db_cache[self.db_cache_key(db_id)] = (backup_data, config_path)

    def _backup_database(self, db_id):
        """Actual backup the current database.

        return a value to be stored in db_cache to allow restoration"""
        raise NotImplementedError()

    def restore_database(self, db_id):
        """Restore a database.

        takes as argument value stored in db_cache by self._backup_database"""
        # XXX set a clearer error message ???
        backup_coordinates, config_path = self.db_cache[self.db_cache_key(db_id)]
        # reload the config used to create the database.
        with open(config_path, "rb") as f:
            config = pickle.load(f)
        # shutdown repo before changing database content
        if self._repo is not None:
            self._repo.turn_repo_off()
        self._restore_database(backup_coordinates, config)

    def _restore_database(self, backup_coordinates, config):
        """Actual restore of the current database.

        Use the value stored in db_cache as input"""
        raise NotImplementedError()

    def get_repo(self, startup=False):
        """return Repository object on the current database.

        (turn the current repo object "on" if there is one or recreate one)
        if startup is True, server startup server hooks will be called if needed
        """
        if self._repo is None:
            self._repo = self._new_repo(self.config)
        # config has now been bootstrapped, call init_config if specified
        if self.init_config is not None:
            self.init_config(self.config)
        repo = self._repo
        repo.turn_repo_on()
        if startup and not repo._has_started:
            repo.hm.call_hooks("server_startup", repo=repo)
            repo._has_started = True
        return repo

    def _new_repo(self, config):
        """Factory method to create a new Repository Instance"""
        config._cubes = None
        repo = config.repository()
        config.repository = lambda vreg=None: repo
        # extending Repository class
        repo._has_started = False
        repo._needs_refresh = False
        repo.turn_repo_on = partial(turn_repo_on, repo)
        repo.turn_repo_off = partial(turn_repo_off, repo)
        return repo

    def get_cnx(self):
        """return Connection object on the current repository"""
        from cubicweb.repoapi import connect

        repo = self.get_repo()
        sources = self.config.read_sources_file()
        login = sources["admin"]["login"]
        password = sources["admin"]["password"] or "xxx"
        cnx = connect(repo, login, password=password)
        return cnx

    def get_repo_and_cnx(self, db_id=DEFAULT_EMPTY_DB_ID):
        """Reset database with the current db_id and return (repo, cnx)

        A database *MUST* have been build with the current <db_id> prior to
        call this method. See the ``build_db_cache`` method. The returned
        repository have it's startup hooks called and the connection is
        establised as admin."""

        self.restore_database(db_id)
        repo = self.get_repo(startup=True)
        cnx = self.get_cnx()
        return repo, cnx

    @property
    def system_source(self):
        return self.config.system_source_config

    @property
    def dbname(self):
        return self.system_source["db-name"]

    def init_test_database(self):
        """actual initialisation of the database"""
        raise ValueError(f"no initialization function for driver {self.DRIVER!r}")

    def has_cache(self, db_id):
        """Check if a given database id exist in db cache for the current config"""
        key = self.db_cache_key(db_id)
        if key in self.db_cache:
            return True
        self.discover_cached_db()
        return key in self.db_cache

    def discover_cached_db(self):
        """Search available db_if for the current config"""
        cache_glob = self.absolute_backup_file("*", "*")
        directory = os.path.dirname(cache_glob)
        entries = {}
        candidates = glob.glob(cache_glob)
        for filepath in candidates:
            data = os.path.basename(filepath)
            # database backup are in the forms are <dbname>-<db_id>.<backtype>
            dbname, data = data.split("-", 1)
            db_id, filetype = data.split(".", 1)
            entries.setdefault((dbname, db_id), {})[filetype] = filepath
        for (dbname, db_id), entry in entries.items():
            # apply necessary transformation from the driver
            value = self.process_cache_entry(directory, dbname, db_id, entry)
            assert "config" in entry
            if value is not None:  # None value means "not handled by this driver"
                key = self.db_cache_key(db_id, dbname=dbname)
                self.db_cache[key] = value, entry["config"]
        self.explored_glob.add(cache_glob)

    def process_cache_entry(self, directory, dbname, db_id, entry):
        """Transforms potential cache entry to proper backup coordinate

        entry argument is a "filetype" -> "filepath" mapping
        Return None if an entry should be ignored."""
        return None

    def build_db_cache(self, test_db_id=DEFAULT_EMPTY_DB_ID, pre_setup_func=None):
        """Build Database cache for ``test_db_id`` if a cache doesn't exist

        if ``test_db_id is DEFAULT_EMPTY_DB_ID`` self.init_test_database is
        called. otherwise, DEFAULT_EMPTY_DB_ID is build/restored and
        ``pre_setup_func`` to setup the database.

        This function backup any database it build"""
        lockfile = join(self._ensure_test_backup_db_dir(), f"{test_db_id}.lock")
        with filelock.FileLock(lockfile):
            if self.has_cache(test_db_id):
                return  # test_db_id, 'already in cache'
            if test_db_id is DEFAULT_EMPTY_DB_ID:
                self.init_test_database()
            else:
                print(f"Building {test_db_id} for database {self.dbname}")
                self.build_db_cache(DEFAULT_EMPTY_DB_ID)
                self.restore_database(DEFAULT_EMPTY_DB_ID)
                self.get_repo(startup=True)
                cnx = self.get_cnx()
                with cnx:
                    pre_setup_func(cnx, self.config)
                    cnx.commit()
            self.backup_database(test_db_id)


class NoCreateDropDatabaseHandler(TestDataBaseHandler):
    """This handler is used if config.skip_db_create_and_restore is True

    This is typically the case with RealDBConfig. In that case,
    we explicitely want to skip init / backup / restore phases.

    This handler redefines the three corresponding methods and delegates
    to original handler for any other method / attribute
    """

    def __init__(self, base_handler):
        self.base_handler = base_handler

    # override init / backup / restore methods
    def init_test_database(self):
        pass

    def backup_database(self, db_id):
        pass

    def restore_database(self, db_id):
        pass

    # delegate to original handler in all other cases
    def __getattr__(self, attrname):
        return getattr(self.base_handler, attrname)


# ## postgres test database handling ############################################


def startpgcluster(pyfile):
    """Start a postgresql cluster next to pyfile"""
    datadir = join(
        os.path.dirname(pyfile),
        "data",
        "database",
        f"pgdb-{os.path.splitext(os.path.basename(pyfile))[0]}",
    )
    if not exists(datadir):
        try:
            subprocess.check_call(
                ["initdb", "-D", datadir, "-E", "utf-8", "--locale=C"]
            )

        except OSError as err:
            if err.errno == errno.ENOENT:
                raise OSError(
                    '"initdb" could not be found. '
                    "You should add the postgresql bin folder to your PATH "
                    "(/usr/lib/postgresql/9.1/bin for example)."
                )
            raise
    datadir = os.path.abspath(datadir)
    pgport = "5432"
    env = os.environ.copy()
    sockdir = tempfile.mkdtemp(prefix="cwpg")
    DEFAULT_PSQL_SOURCES["system"]["db-host"] = sockdir
    DEFAULT_PSQL_SOURCES["system"]["db-port"] = pgport
    options = f'-h "" -k {sockdir} -p {pgport}'
    options += " -c fsync=off -c full_page_writes=off"
    options += " -c synchronous_commit=off"

    command = ["pg_ctl", "start", "-w", "-D", datadir, "-o", options]
    try:
        # don't use check_output or run() here, it will deadlock
        p = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env
        )
        p.wait()

        if p.returncode != 0:
            printable_command = " ".join(shlex.quote(arg) for arg in command)
            raise Exception(
                f'"pg_ctl" command failed with the return code '
                f'"{p.returncode}".\n\nFull command:\n\n    {printable_command}\n\n'
                f'Output:\n\n{p.stdout.read().decode("Utf-8")}'
            )

    except OSError as err:
        try:
            os.rmdir(sockdir)
        except OSError:
            pass
        if err.errno == errno.ENOENT:
            raise OSError(
                '"pg_ctl" could not be found. '
                "You should add the postgresql bin folder to your PATH "
                "(/usr/lib/postgresql/9.1/bin for example)."
            )
        raise


def stoppgcluster(pyfile):
    """Kill the postgresql cluster running next to pyfile"""
    datadir = join(
        os.path.dirname(pyfile),
        "data",
        "database",
        f"pgdb-{os.path.splitext(os.path.basename(pyfile))[0]}",
    )
    subprocess.call(["pg_ctl", "stop", "-D", datadir, "-m", "fast"])
    try:
        os.rmdir(DEFAULT_PSQL_SOURCES["system"]["db-host"])
    except OSError:
        pass


class PostgresTestDataBaseHandler(TestDataBaseHandler):
    DRIVER = "postgres"

    # Separate db_cache for PG databases, to avoid collisions with sqlite dbs
    db_cache = {}
    explored_glob = set()

    __CTL = set()

    def __init__(self, *args, **kwargs):
        super(PostgresTestDataBaseHandler, self).__init__(*args, **kwargs)
        if "global-db-name" not in self.system_source:
            self.system_source["global-db-name"] = self.system_source["db-name"]
            self.system_source["db-name"] = self.system_source["db-name"] + str(
                os.getpid()
            )

    @property
    @cached
    def helper(self):
        from logilab.database import get_db_helper

        return get_db_helper("postgres")

    @property
    def dbname(self):
        return self.system_source["global-db-name"]

    @property
    def dbcnx(self):
        try:
            return self._cnx
        except AttributeError:
            from cubicweb.server.serverctl import _db_sys_cnx

            try:
                self._cnx = _db_sys_cnx(
                    self.system_source,
                    "CREATE DATABASE and / or USER",
                    interactive=False,
                )
                return self._cnx
            except Exception:
                self._cnx = None
                raise

    @property
    @cached
    def cursor(self):
        return self.dbcnx.cursor()

    def process_cache_entry(self, directory, dbname, db_id, entry):
        backup_name = self._backup_name(db_id)
        if backup_name in self.helper.list_databases(self.cursor):
            return backup_name
        return None

    def has_cache(self, db_id):
        backup_name = self._backup_name(db_id)
        return super(PostgresTestDataBaseHandler, self).has_cache(
            db_id
        ) and backup_name in self.helper.list_databases(self.cursor)

    def init_test_database(self):
        """initialize a fresh postgresql database used for testing purpose"""
        from cubicweb.server import init_repository
        from cubicweb.server.serverctl import source_cnx, createdb

        # connect on the dbms system base to create our base
        try:
            self._drop(self.system_source["db-name"])
            createdb(self.helper, self.system_source, self.dbcnx, self.cursor)
            self.dbcnx.commit()
            cnx = source_cnx(
                self.system_source, special_privs="LANGUAGE C", interactive=False
            )
            templcursor = cnx.cursor()
            try:
                # XXX factorize with db-create code
                self.helper.init_fti_extensions(templcursor)
                # install plpythonu/plpgsql language if not installed by the cube
                langs = ("plpgsql",)
                for extlang in langs:
                    self.helper.create_language(templcursor, extlang)
                cnx.commit()
            finally:
                templcursor.close()
                cnx.close()
            init_repository(
                self.config, interactive=False, init_config=self.init_config
            )
        except BaseException:
            if self.dbcnx is not None:
                self.dbcnx.rollback()
            sys.stderr.write(f"building {self.dbname} failed\n")
            raise

    def helper_clear_cache(self):
        # clear_cache might be unavailable during process termination, see:
        # https://docs.python.org/3/reference/datamodel.html#object.__del__

        if self.dbcnx is not None:
            self.dbcnx.commit()
            self.dbcnx.close()
            del self._cnx
            if clear_cache is not None:
                clear_cache(self, "cursor")

        if clear_cache is not None:
            clear_cache(self, "helper")

    def __del__(self):
        self.helper_clear_cache()

    @property
    def _config_id(self):
        return sha1(self.config.apphome.encode("utf-8")).hexdigest()[:10]

    def _backup_name(self, db_id):  # merge me with parent
        backup_name = "_".join(("cache", self._config_id, self.dbname, db_id))
        return backup_name.lower()

    def _drop(self, db_name):
        if db_name in self.helper.list_databases(self.cursor):
            self.cursor.execute(f"DROP DATABASE {db_name}")
            self.dbcnx.commit()

    def _backup_database(self, db_id):
        """Actual backup the current database.

        return a value to be stored in db_cache to allow restoration
        """
        from cubicweb.server.serverctl import createdb

        orig_name = self.system_source["db-name"]
        try:
            backup_name = self._backup_name(db_id)
            self._drop(backup_name)
            self.system_source["db-name"] = backup_name
            if self._repo:
                self._repo.turn_repo_off()
            try:
                createdb(
                    self.helper,
                    self.system_source,
                    self.dbcnx,
                    self.cursor,
                    template=orig_name,
                )
                self.dbcnx.commit()
            finally:
                if self._repo:
                    self._repo.turn_repo_on()
            return backup_name
        finally:
            self.system_source["db-name"] = orig_name

    def _restore_database(self, backup_coordinates, config):
        from cubicweb.server.serverctl import createdb

        """Actual restore of the current database.

        Use the value tostored in db_cache as input """
        self._drop(self.system_source["db-name"])
        createdb(
            self.helper,
            self.system_source,
            self.dbcnx,
            self.cursor,
            template=backup_coordinates,
        )
        self.dbcnx.commit()


# sqlite test database handling ################################################


class SQLiteTestDataBaseHandler(TestDataBaseHandler):
    DRIVER = "sqlite"

    __TMPDB = set()

    @classmethod
    def _cleanup_all_tmpdb(cls):
        for dbpath in cls.__TMPDB:
            cls._cleanup_database(dbpath)

    def __init__(self, *args, **kwargs):
        super(SQLiteTestDataBaseHandler, self).__init__(*args, **kwargs)
        # use a dedicated base for each process.
        if "global-db-name" not in self.system_source:
            self.system_source["global-db-name"] = self.system_source["db-name"]
            process_db = self.system_source["db-name"] + str(os.getpid())
            self.system_source["db-name"] = process_db
        process_db = self.absolute_dbfile()  # update db-name to absolute path
        self.__TMPDB.add(process_db)

    @staticmethod
    def _cleanup_database(dbfile):
        try:
            os.remove(dbfile)
            os.remove(f"{dbfile}-journal")
        except OSError:
            pass

    @property
    def dbname(self):
        return self.system_source["global-db-name"]

    def absolute_dbfile(self):
        """absolute path of current database file"""
        dbfile = join(self._ensure_test_backup_db_dir(), self.system_source["db-name"])
        self.system_source["db-name"] = dbfile
        return dbfile

    def process_cache_entry(self, directory, dbname, db_id, entry):
        return entry.get("sqlite")

    def _backup_database(self, db_id=DEFAULT_EMPTY_DB_ID):
        # XXX remove database file if it exists ???
        dbfile = self.absolute_dbfile()
        backup_file = self.absolute_backup_file(db_id, "sqlite")
        shutil.copy(dbfile, backup_file)
        return backup_file

    def _restore_database(self, backup_coordinates, _config):
        # remove database file if it exists ?
        dbfile = self.absolute_dbfile()
        self._cleanup_database(dbfile)
        shutil.copy(backup_coordinates, dbfile)
        self.get_repo()

    def init_test_database(self):
        """initialize a fresh sqlite databse used for testing purpose"""
        # initialize the database
        from cubicweb.server import init_repository

        self._cleanup_database(self.absolute_dbfile())
        init_repository(self.config, interactive=False, init_config=self.init_config)


atexit.register(SQLiteTestDataBaseHandler._cleanup_all_tmpdb)


HANDLERS = {}


def register_handler(handlerkls, overwrite=False):
    assert handlerkls is not None
    if overwrite or handlerkls.DRIVER not in HANDLERS:
        HANDLERS[handlerkls.DRIVER] = handlerkls
    else:
        msg = (
            "%s: Handler already exists use overwrite if it's intended\n"
            "(existing handler class is %r)"
        )
        raise ValueError(msg % (handlerkls.DRIVER, HANDLERS[handlerkls.DRIVER]))


register_handler(PostgresTestDataBaseHandler)
register_handler(SQLiteTestDataBaseHandler)


class HCache:
    """Handler cache object: store database handler for a given configuration.

    We only keep one repo in cache to prevent too much objects to stay alive
    (database handler holds a reference to a repository). As at the moment a new
    handler is created for each TestCase class and all test methods are executed
    sequentially whithin this class, there should not have more cache miss that
    if we had a wider cache as once a Handler stop being used it won't be used
    again.
    """

    def __init__(self):
        self.config = None
        self.handler = None

    def get(self, config):
        if config is self.config:
            return self.handler
        else:
            return None

    def set(self, config, handler):
        self.config = config
        self.handler = handler


HCACHE = HCache()

# XXX a class method on Test ?
_CONFIG = None


def get_test_db_handler(config, init_config=None):
    global _CONFIG
    if _CONFIG is not None and config is not _CONFIG:
        from logilab.common.modutils import cleanup_sys_modules

        # cleanup all dynamically loaded modules and everything in the instance
        # directory
        apphome = _CONFIG.apphome
        if apphome:  # may be unset in tests
            cleanup_sys_modules([apphome])
        # also cleanup sys.path
        if apphome in sys.path:
            sys.path.remove(apphome)
    _CONFIG = config
    config.adjust_sys_path()
    handler = HCACHE.get(config)
    if handler is not None:
        return handler
    driver = config.system_source_config["db-driver"]
    handlerkls = HANDLERS.get(driver, None)
    if handlerkls is not None:
        handler = handlerkls(config, init_config)
        if config.skip_db_create_and_restore:
            handler = NoCreateDropDatabaseHandler(handler)
        HCACHE.set(config, handler)
        return handler
    else:
        raise ValueError(f"no initialization function for driver {driver!r}")



from os import unlink
from os.path import isfile, join

from cubicweb.cwconfig import CubicWebConfiguration as cwcfg

regdir = cwcfg.instances_dir()

if isfile(join(regdir, "startorder")):
    if confirm(
        "The startorder file is not used anymore in Cubicweb 3.22. "
        "Should I delete it?",
        shell=False,
        pdb=False,
    ):
        unlink(join(regdir, "startorder"))



# pylint: disable-msg=W0622
"""cubicweb-fakeemail packaging information"""

from glob import glob
from os import listdir as _listdir
from os.path import join, isdir

modname = "fakecustomtype"
distname = f"cubicweb-{modname}"

numversion = (1, 0, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "Logilab"
author_email = "contact@logilab.fr"
web = f"http://www.cubicweb.org/project/{distname}"
description = "whatever"
classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: JavaScript",
]

# used packages
__depends__ = {
    "cubicweb": ">= 3.19.0",
}


# packaging ###


THIS_CUBE_DIR = join("share", "cubicweb", "cubes", modname)


def listdir(dirpath):
    return [
        join(dirpath, fname)
        for fname in _listdir(dirpath)
        if fname[0] != "."
        and not fname.endswith(".pyc")
        and not fname.endswith("~")
        and not isdir(join(dirpath, fname))
    ]


data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob("*.py") if fname != "setup.py"]],
]
# check for possible extended cube layout
for dirname in (
    "entities",
    "views",
    "sobjects",
    "hooks",
    "schema",
    "data",
    "i18n",
    "migration",
    "wdoc",
):
    if isdir(dirname):
        data_files.append([join(THIS_CUBE_DIR, dirname), listdir(dirname)])



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Workflow setup utilities.

These functions work with a declarative workflow definition:

.. code-block:: python

        {
            'etypes': 'CWGroup',
            'default': True,
            'initial_state': u'draft',
            'states': [u'draft', u'published'],
            'transitions': {
                u'publish': {
                    'fromstates': u'draft',
                    'tostate': u'published',
                    'requiredgroups': u'managers'
                    'conditions': (
                        'U in_group X',
                        'X owned_by U'
                    )
                }
            }
        }

.. autofunction:: setup_workflow
.. autofunction:: cleanupworkflow
"""

import collections

from cubicweb import NoResultError


def get_tuple_or_list(value):
    if value is None:
        return None
    if not isinstance(value, (tuple, list)):
        value = (value,)
    return value


def cleanupworkflow(cnx, wf, wfdef):
    """Cleanup an existing workflow by removing the states and transitions that
    do not exist in the given definition.

    :param cnx: A connexion with enough permissions to define a workflow
    :param wf: A `Workflow` entity
    :param wfdef: A workflow definition
    """
    cnx.execute(
        "DELETE State S WHERE S state_of WF, WF eid %%(wf)s, "
        "NOT S name IN (%s)" % (", ".join('"%s"' % s for s in wfdef["states"])),
        {"wf": wf.eid},
    )

    cnx.execute(
        "DELETE Transition T WHERE T transition_of WF, WF eid %%(wf)s, "
        "NOT T name IN (%s)" % (", ".join('"%s"' % s for s in wfdef["transitions"])),
        {"wf": wf.eid},
    )


def setup_workflow(cnx, name, wfdef, cleanup=True):
    """Create or update a workflow definition so it matches the given
    definition.

    :param cnx: A connexion with enough permissions to define a workflow
    :param name: The workflow name. Used to create the `Workflow` entity, or
                 to find an existing one.
    :param wfdef: A workflow definition.
    :param cleanup: Remove extra states and transitions. Can be done separatly
                    by calling :func:`cleanupworkflow`.
    :return: The created/updated workflow entity
    """
    try:
        wf = cnx.find("Workflow", name=name).one()
    except NoResultError:
        wf = cnx.create_entity("Workflow", name=name)

    etypes = get_tuple_or_list(wfdef["etypes"])
    cnx.execute(
        "DELETE WF workflow_of ETYPE WHERE WF eid %%(wf)s, "
        "NOT ETYPE name IN (%s)" % ",".join('"%s"' for e in etypes),
        {"wf": wf.eid},
    )
    cnx.execute(
        "SET WF workflow_of ETYPE WHERE"
        " NOT WF workflow_of ETYPE, WF eid %%(wf)s, ETYPE name IN (%s)"
        % ",".join('"%s"' % e for e in etypes),
        {"wf": wf.eid},
    )
    if wfdef["default"]:
        cnx.execute(
            "SET ETYPE default_workflow X "
            "WHERE "
            "NOT ETYPE default_workflow X, "
            "X eid %%(x)s, ETYPE name IN (%s)" % ",".join('"%s"' % e for e in etypes),
            {"x": wf.eid},
        )

    states = {}
    states_transitions = collections.defaultdict(list)
    for state in wfdef["states"]:
        st = wf.state_by_name(state) or wf.add_state(state)
        states[state] = st

    if "initial_state" in wfdef:
        wf.cw_set(initial_state=states[wfdef["initial_state"]])

    for trname, trdef in wfdef["transitions"].items():
        tr = wf.transition_by_name(trname) or cnx.create_entity(
            "Transition", name=trname
        )
        tr.cw_set(transition_of=wf)
        if trdef.get("tostate"):
            tr.cw_set(destination_state=states[trdef["tostate"]])
        fromstates = get_tuple_or_list(trdef.get("fromstates", ()))
        for stname in fromstates:
            states_transitions[stname].append(tr)

        requiredgroups = get_tuple_or_list(trdef.get("requiredgroups", ()))
        conditions = get_tuple_or_list(trdef.get("conditions", ()))

        tr.set_permissions(requiredgroups, conditions, reset=True)

    for stname, transitions in states_transitions.items():
        state = states[stname]
        state.cw_set(allowed_transition=None)
        state.cw_set(allowed_transition=transitions)

    if cleanup:
        cleanupworkflow(cnx, wf, wfdef)

    return wf



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""functional tests for fast_drop_entities"""

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.misc.scripts.fast_drop_entities import (
    fast_drop_entities,
    _compute_impacted_entities_and_relations_to_delete,
)


class FastDropEntitiesTC(CubicWebTC):
    def test_simple_deletion(self):
        with self.admin_access.repo_cnx() as cnx:
            for letter in "abc":
                cnx.execute(
                    "INSERT Ami X: X nom %(n)s, X prenom %(p)s",
                    {"n": letter.upper(), "p": letter},
                )

            cnx.commit()

            self.assertEqual(len(cnx.find("Ami")), 3)

            fast_drop_entities(cnx.find("Ami"))

            self.assertEqual(len(cnx.find("Ami")), 0)

    def test_deletion_with_inlined_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            u = cnx.create_entity("Usine", lieu="Paris")
            p = cnx.create_entity("Produit", fabrique_par=u)
            cnx.commit()

            self.assertEqual(p.fabrique_par, (u,))

            fast_drop_entities(cnx.find("Usine"))

            self.assertEqual(len(cnx.find("Usine")), 0)
            self.assertEqual(cnx.find("Produit").one().fabrique_par, ())

    def test_deletion_with_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            alfred = cnx.create_entity("Personne", nom="Alfred")
            tag = cnx.create_entity("Tag", name="A Tag", tags=alfred)
            cnx.commit()

            self.assertEqual(alfred.reverse_tags, (tag,))

            fast_drop_entities(cnx.find("Personne"))

            self.assertEqual(len(cnx.find("Personne")), 0)
            self.assertEqual(cnx.find("Tag", name="A Tag").one().tags, ())

    def test_deletion_with_composites(self):
        with self.admin_access.repo_cnx() as cnx:
            abc = cnx.create_entity("Ami", nom="ABC")
            comment_ABC = cnx.create_entity("Comment", content="ABC", comments=abc)
            comment_about_ABC = cnx.create_entity(
                "Comment", content="I would say BCD instead", comments=comment_ABC
            )
            cnx.commit()

            self.assertEqual(comment_ABC.comments, (abc,))
            self.assertEqual(comment_about_ABC.comments, (comment_ABC,))

            fast_drop_entities(cnx.find("Ami"))

            self.assertEqual(len(cnx.find("Comment")), 0)

    def test_expected_computed_impacted_entities_and_relations(self):
        with self.admin_access.repo_cnx() as cnx:
            abc = cnx.create_entity("Ami", nom="ABC")
            cnx.create_entity("Tag", name="A Tag", tags=abc)
            comment = cnx.create_entity("Comment", content="ABC", comments=abc)
            cnx.commit()

            to_delete = {}
            rels_to_delete = {}
            to_update = {}
            _compute_impacted_entities_and_relations_to_delete(
                cnx, abc.eid, abc.cw_etype, to_delete, rels_to_delete, to_update
            )

            self.assertEqual(to_delete, {"Ami": {abc.eid}, "Comment": {comment.eid}})

            self.assertEqual(to_update, {})  # FIXME: Populate to_update

            self.assertEqual(
                rels_to_delete,
                {
                    ("is", "subject"): {
                        abc.eid,
                        comment.eid,
                    },
                    ("is_instance_of", "subject"): {
                        abc.eid,
                        comment.eid,
                    },
                    ("cw_source", "subject"): {
                        abc.eid,
                        comment.eid,
                    },
                    ("owned_by", "subject"): {
                        abc.eid,
                        comment.eid,
                    },
                    ("created_by", "subject"): {
                        abc.eid,
                        comment.eid,
                    },
                    ("tags", "object"): {
                        abc.eid,
                    },
                },
            )



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Base classes and utilities for cubicweb tests"""

import sys
from contextlib import contextmanager
from inspect import isgeneratorfunction
from itertools import chain
from os.path import dirname, join, abspath
from unittest import TestCase

import yams.schema
from logilab.common.debugger import Debugger
from logilab.common.decorators import cached, classproperty, clear_cache, iclassmethod
from logilab.common.testlib import Tags
from logilab.common.textutils import unormalize

from cubicweb import (
    AuthenticationError,
    BadConnectionId,
)
from cubicweb import devtools, repoapi, server
from cubicweb.devtools import (
    SYSTEM_ENTITIES,
)
from cubicweb.devtools import DEFAULT_EMPTY_DB_ID
from cubicweb.devtools.apptest_config import (
    ApptestConfiguration,
    MAILBOX,
)
from cubicweb.server.hook import SendMailOp
from cubicweb.server.session import Connection
from cubicweb.utils import json, make_uid


# provide a data directory for the test class ##################################


class BaseTestCase(TestCase):
    @classproperty
    @cached
    def datadir(cls):  # pylint: disable=E0213
        """helper attribute holding the standard test's data directory"""
        mod = sys.modules[cls.__module__]
        return join(dirname(abspath(mod.__file__)), "data")

    # cache it (use a class method to cache on class since TestCase is
    # instantiated for each test run)

    @classmethod
    def datapath(cls, *fname):
        """joins the object's datadir and `fname`"""
        return join(cls.datadir, *fname)


# low-level utilities ##########################################################


class CubicWebDebugger(Debugger):
    """special debugger class providing a 'view' function which saves some
    html into a temporary file and open a web browser to examinate it.
    """

    def do_view(self, arg):
        import webbrowser

        data = self._getval(arg)
        with open("/tmp/toto.html", "w") as toto:
            toto.write(data)
        webbrowser.open("file:///tmp/toto.html")


def line_context_filter(line_no, center, before=3, after=None):
    """return true if line are in context

    if after is None: after = before
    """
    if after is None:
        after = before
    return center - before <= line_no <= center + after


def unprotected_entities(schema, strict=False):
    """returned a set of each non final entity type, excluding "system" entities
    (eg CWGroup, CWUser...)
    """
    if strict:
        protected_entities = yams.schema.BASE_TYPES
    else:
        protected_entities = yams.schema.BASE_TYPES.union(SYSTEM_ENTITIES)
    return set(schema.entities()) - protected_entities


class JsonValidator:
    def parse_string(self, data):
        return json.loads(data.decode("ascii"))


@contextmanager
def real_error_handling(app):
    """By default, CubicWebTC `app` attribute (ie the publisher) is monkey
    patched so that unexpected error are raised rather than going through the
    `error_handler` method.

    By using this context manager you disable this monkey-patching temporarily.
    Hence when publishihng a request no error will be raised, you'll get
    req.status_out set to an HTTP error status code and the generated page will
    usually hold a traceback as HTML.

    >>> with real_error_handling(app):
    >>>     page = app.handle_request(req)
    """
    # remove the monkey patched error handler
    fake_error_handler = app.error_handler
    del app.error_handler
    # return the app
    yield app
    # restore
    app.error_handler = fake_error_handler


# Repoaccess utility ###############################################3###########


class Session:
    """In-memory user session"""

    def __init__(self, repo, user):
        self.user = user  # XXX deprecate and store only a login.
        self.repo = repo
        self.sessionid = make_uid(unormalize(user.login))
        self.data = {}

    def __str__(self):
        return f"<session {self.user.login} (0x{id(self):x})>"

    @property
    def anonymous_session(self):
        # XXX for now, anonymous_user only exists in webconfig (and testconfig).
        # It will only be present inside all-in-one instance.
        # there is plan to move it down to global config.
        if not hasattr(self.repo.config, "anonymous_user"):
            # not a web or test config, no anonymous user
            return False
        return self.user.login == self.repo.config.anonymous_user()[0]

    def new_cnx(self):
        """Return a new Connection object linked to the session

        The returned Connection will *not* be managed by the Session.
        """
        cnx = Connection(self.repo, self.user)
        cnx.session = self
        return cnx


class RepoAccess:
    """An helper to easily create object to access the repo as a specific user

    Each RepoAccess have it own session.

    A repo access can create three type of object:

    .. automethod:: cubicweb.testlib.RepoAccess.cnx
    .. automethod:: cubicweb.testlib.RepoAccess.web_request
    """

    def __init__(self, repo, login):
        self._repo = repo
        self._login = login
        with repo.internal_cnx() as cnx:
            self._user = cnx.find("CWUser", login=login).one()
            self._user.cw_attr_cache["login"] = login

    @contextmanager
    def cnx(self):
        """Context manager returning a server side connection for the user"""
        with repoapi.Connection(self._repo, self._user) as cnx:
            yield cnx

    # aliases for bw compat
    client_cnx = repo_cnx = cnx

    @contextmanager
    def shell(self):
        from cubicweb.server.migractions import ServerMigrationHelper

        with self.cnx() as cnx:
            mih = ServerMigrationHelper(
                None,
                repo=self._repo,
                cnx=cnx,
                interactive=False,
                # hack so it don't try to load fs schema
                schema=1,
            )
            yield mih
            cnx.commit()


# base class for cubicweb tests requiring a full cw environments ###############


class CubicWebTC(BaseTestCase):
    """abstract class for test using an apptest environment

    attributes:

    * `vreg`, the vregistry
    * `schema`, self.vreg.schema
    * `config`, cubicweb configuration
    * `cnx`, repoapi connection to the repository using an admin user
    * `session`, server side session associated to `cnx`
    * `app`, the cubicweb publisher (for web testing)
    * `repo`, the repository object
    * `admlogin`, login of the admin user
    * `admpassword`, password of the admin user
    * `shell`, create and use shell environment
    * `anonymous_allowed`: flag telling if anonymous browsing should be allowed
    """

    appid = "data"
    configcls = ApptestConfiguration
    tags = Tags("cubicweb", "cw_repo")
    test_db_id = DEFAULT_EMPTY_DB_ID

    # anonymous is logged by default in cubicweb test cases
    anonymous_allowed = True

    @classmethod
    def setUpClass(cls):
        test_module_file = sys.modules[cls.__module__].__file__
        assert "config" not in cls.__dict__, (
            "%s has a config class attribute before entering setUpClass. "
            "Let CubicWebTC.setUpClass instantiate it and modify it afterwards." % cls
        )
        cls.config = cls.configcls(cls.appid, test_module_file)
        cls.config.mode = "test"

    def __init__(self, *args, **kwargs):
        self.repo = None
        self._open_access = set()
        super(CubicWebTC, self).__init__(*args, **kwargs)

    def run(self, *args, **kwds):
        testMethod = getattr(self, self._testMethodName)
        if isgeneratorfunction(testMethod):
            raise RuntimeError(
                "%s appears to be a generative test. This is not handled "
                "anymore, use subTest API instead." % self
            )
        return super(CubicWebTC, self).run(*args, **kwds)

    # repository connection handling ###########################################

    def new_access(self, login):
        """provide a new RepoAccess object for a given user

        The access is automatically closed at the end of the test."""
        access = RepoAccess(self.repo, login)
        self._open_access.add(access)
        return access

    def _close_access(self):
        while self._open_access:
            try:
                self._open_access.pop()
            except BadConnectionId:
                continue  # already closed

    def _init_repo(self):
        """init the repository and connection to it."""
        # get or restore and working db.
        db_handler = devtools.get_test_db_handler(self.config, self.init_config)
        db_handler.build_db_cache(self.test_db_id, self.pre_setup_database)
        db_handler.restore_database(self.test_db_id)
        self.repo = db_handler.get_repo(startup=True)
        # get an admin session (without actual login)
        login = db_handler.config.default_admin_config["login"]
        self.admin_access = self.new_access(login)

    # config management ########################################################

    @classmethod  # XXX could be turned into a regular method
    def init_config(cls, config):
        """configuration initialization hooks.

        You may only want to override here the configuraton logic.

        Otherwise, consider to use a different :class:`ApptestConfiguration`
        defined in the `configcls` class attribute.

        This method will be called by the database handler once the config has
        been properly bootstrapped.
        """
        admincfg = config.default_admin_config
        cls.admlogin = admincfg["login"]
        cls.admpassword = admincfg["password"]
        # uncomment the line below if you want rql queries to be logged
        # config.global_set_option('query-log-file',
        #                          '/tmp/test_rql_log.' + `os.getpid()`)
        config.global_set_option("log-file", None)
        # set default-dest-addrs to a dumb email address to avoid mailbox or
        # mail queue pollution
        config.global_set_option("default-dest-addrs", ["whatever"])
        config.global_set_option("sender-addr", "cubicweb-send@logilab.fr")
        config.global_set_option(
            "default-dest-addrs", "cubicweb-default-dest@logilab.fr"
        )
        config.global_set_option("sender-name", "cubicweb-test")
        config.global_set_option("sender-addr", "cubicweb-test@logilab.fr")
        # default_base_url on config class isn't enough for TestServerConfiguration
        config.global_set_option("base-url", config.default_base_url())

    @property
    def vreg(self):
        return self.repo.vreg

    # global resources accessors ###############################################

    @property
    def schema(self):
        """return the application schema"""
        return self.vreg.schema

    def set_option(self, optname, value):
        self.config.global_set_option(optname, value)

    def set_debug(self, debugmode):
        server.set_debug(debugmode)

    def debugged(self, debugmode):
        return server.debugged(debugmode)

    # default test setup and teardown #########################################

    def setUp(self):
        assert hasattr(self, "config"), (
            "It seems that CubicWebTC.setUpClass has not been called. "
            "Missing super() call in %s?" % self.setUpClass
        )
        self.config.set_anonymous_allowed(self.anonymous_allowed)
        # monkey patch send mail operation so emails are sent synchronously
        self._patch_SendMailOp()
        previous_failure = self.__class__.__dict__.get("_repo_init_failed")
        if previous_failure is not None:
            self.skipTest(f"repository is not initialised: {previous_failure!r}")
        try:
            self._init_repo()
        except Exception as ex:
            self.__class__._repo_init_failed = ex
            raise
        self.addCleanup(self._close_access)
        self.setup_database()
        MAILBOX[:] = []  # reset mailbox

        self.previous_log_threshold = self.config.config.log_threshold

        # I want debug all the time because god
        self.config.global_set_option("log-threshold", "DEBUG")
        self.config.init_log(self.config["log-threshold"], force=True)

    def tearDown(self):
        self.config.global_set_option("log-threshold", self.previous_log_threshold)
        self.config.init_log(self.config["log-threshold"], force=True)

        while self._cleanups:
            cleanup, args, kwargs = self._cleanups.pop(-1)
            cleanup(*args, **kwargs)
        self.repo.turn_repo_off()

    def _patch_SendMailOp(self):
        # monkey patch send mail operation so emails are sent synchronously
        _old_mail_postcommit_event = SendMailOp.postcommit_event
        SendMailOp.postcommit_event = SendMailOp.sendmails

        def reverse_SendMailOp_monkey_patch():
            SendMailOp.postcommit_event = _old_mail_postcommit_event

        self.addCleanup(reverse_SendMailOp_monkey_patch)

    def setup_database(self):
        """add your database setup code by overriding this method"""

    @classmethod
    def pre_setup_database(cls, cnx, config):
        """add your pre database setup code by overriding this method

        Do not forget to set the cls.test_db_id value to enable caching of the
        result.
        """

    # user / session management ###############################################

    @iclassmethod  # XXX turn into a class method
    def create_user(
        self,
        req,
        login=None,
        groups=("users",),
        password=None,
        email=None,
        commit=True,
        **kwargs,
    ):
        """create and return a new user entity"""
        if password is None:
            password = login
        user = req.create_entity("CWUser", login=login, upassword=password, **kwargs)
        req.execute(
            "SET X in_group G WHERE X eid %%(x)s, G name IN(%s)"
            % ",".join(repr(str(g)) for g in groups),
            {"x": user.eid},
        )
        if email is not None:
            req.create_entity("EmailAddress", address=email, reverse_primary_email=user)
        user.cw_clear_relation_cache("in_group", "subject")
        if commit:
            getattr(req, "cnx", req).commit()
        return user

    # other utilities #########################################################

    @contextmanager
    def temporary_appobjects(self, *appobjects):
        self.vreg._loadedmods.setdefault(self.__module__, {})
        for obj in appobjects:
            self.vreg.register(obj)
            registered = getattr(obj, "__registered__", None)
            if registered:
                for registry in obj.__registries__:
                    registered(self.vreg[registry])
        try:
            yield
        finally:
            for obj in appobjects:
                self.vreg.unregister(obj)

    @contextmanager
    def temporary_permissions(self, *perm_overrides, **perm_kwoverrides):
        """Set custom schema permissions within context.

        There are two ways to call this method, which may be used together :

        * using positional argument(s):

          .. sourcecode:: python

                rdef = self.schema['CWUser'].rdef('login')
                with self.temporary_permissions((rdef, {'read': ()})):
                    ...


        * using named argument(s):

          .. sourcecode:: python

                with self.temporary_permissions(CWUser={'read': ()}):
                    ...

        Usually the former will be preferred to override permissions on a
        relation definition, while the latter is well suited for entity types.

        The allowed keys in the permission dictionary depend on the schema type
        (entity type / relation definition). Resulting permissions will be
        similar to `orig_permissions.update(partial_perms)`.
        """
        torestore = []
        for erschema, etypeperms in chain(perm_overrides, perm_kwoverrides.items()):
            if isinstance(erschema, str):
                erschema = self.schema[erschema]
            for action, actionperms in etypeperms.items():
                origperms = erschema.permissions[action]
                erschema.set_action_permissions(action, actionperms)
                torestore.append([erschema, action, origperms])
        try:
            yield
        finally:
            for erschema, action, permissions in torestore:
                if action is None:
                    erschema.permissions = permissions
                else:
                    erschema.set_action_permissions(action, permissions)

    def assertModificationDateGreater(self, entity, olddate):
        entity.cw_attr_cache.pop("modification_date", None)
        self.assertGreater(entity.modification_date, olddate)

    def assertMessageEqual(self, req, params, expected_msg):
        msg = req.session.data[params["_cwmsgid"]]
        self.assertEqual(expected_msg, msg)

    def assertPossibleTransitions(self, entity, expected):
        transitions = entity.cw_adapt_to("IWorkflowable").possible_transitions()
        self.assertListEqual(sorted(tr.name for tr in transitions), sorted(expected))

    def set_auth_mode(self, authmode, anonuser=None):
        self.set_option("auth-mode", authmode)
        self.set_option("anonymous-user", anonuser)
        if anonuser is None:
            self.config.anonymous_credential = None
        else:
            self.config.anonymous_credential = (anonuser, anonuser)

    def init_authentication(self, authmode, anonuser=None):
        self.set_auth_mode(authmode, anonuser)
        req = self.requestcls(self.vreg, url="login")
        sh = self.app.session_handler
        authm = sh.session_manager.authmanager
        authm.anoninfo = self.vreg.config.anonymous_user()
        authm.anoninfo = authm.anoninfo[0], {"password": authm.anoninfo[1]}
        # not properly cleaned between tests
        self.open_sessions = sh.session_manager._sessions = {}
        return req

    def assertAuthSuccess(self, req, nbsessions=1):
        session = self.app.get_session(req)
        cnx = session.new_cnx()
        with cnx:
            req.set_cnx(cnx)
        self.assertEqual(len(self.open_sessions), nbsessions, self.open_sessions)
        self.assertEqual(req.user.login, self.admlogin)
        self.assertEqual(session.anonymous_session, False)

    def assertAuthFailure(self, req, nbsessions=0):
        with self.assertRaises(AuthenticationError):
            self.app.get_session(req)
        # +0 since we do not track the opened session
        self.assertEqual(len(self.open_sessions), nbsessions)
        clear_cache(req, "get_authorization")

    # notifications ############################################################

    def assertSentEmail(self, subject, recipients=None, nb_msgs=None):
        """test recipients in system mailbox for given email subject

        :param subject: email subject to find in mailbox
        :param recipients: list of email recipients
        :param nb_msgs: expected number of entries
        :returns: list of matched emails
        """
        messages = [
            email for email in MAILBOX if email.message.get("Subject") == subject
        ]
        if recipients is not None:
            sent_to = set()
            for msg in messages:
                sent_to.update(msg.recipients)
            self.assertSetEqual(set(recipients), sent_to)
        if nb_msgs is not None:
            self.assertEqual(len(MAILBOX), nb_msgs)
        return messages


# registry instrumentization ###################################################


def not_selected(vreg, appobject):
    try:
        vreg._selected[appobject.__class__] -= 1
    except (KeyError, AttributeError):
        pass



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Core hooks: workflow related hooks"""


from datetime import datetime, timezone

from cubicweb import RepositoryError, validation_error, _
from cubicweb.predicates import is_instance, adaptable
from cubicweb.server import hook


def _change_state(cnx, x, oldstate, newstate):
    nocheck = cnx.transaction_data.setdefault("skip-security", set())
    nocheck.add((x, "in_state", oldstate))
    nocheck.add((x, "in_state", newstate))
    # delete previous state first
    cnx.delete_relation(x, "in_state", oldstate)
    cnx.add_relation(x, "in_state", newstate)


# operations ###################################################################


class _SetInitialStateOp(hook.Operation):
    """make initial state be a default state"""

    eid = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        entity = cnx.entity_from_eid(self.eid)
        iworkflowable = entity.cw_adapt_to("IWorkflowable")
        # if there is an initial state and the entity's state is not set,
        # use the initial state as a default state
        if (
            not (cnx.deleted_in_transaction(entity.eid) or entity.in_state)
            and iworkflowable.current_workflow
        ):
            state = iworkflowable.current_workflow.initial
            if state:
                cnx.add_relation(self.eid, "in_state", state.eid)
                _FireAutotransitionOp(cnx, eid=self.eid)


class _FireAutotransitionOp(hook.Operation):
    """try to fire auto transition after state changes"""

    eid = None  # make pylint happy

    def precommit_event(self):
        entity = self.cnx.entity_from_eid(self.eid)
        iworkflowable = entity.cw_adapt_to("IWorkflowable")
        autotrs = list(iworkflowable.possible_transitions("auto"))
        if autotrs:
            assert len(autotrs) == 1
            iworkflowable.fire_transition(autotrs[0])


class _WorkflowChangedOp(hook.Operation):
    """fix entity current state when changing its workflow"""

    eid = wfeid = None  # make pylint happy

    def precommit_event(self):
        # notice that enforcement that new workflow apply to the entity's type is
        # done by schema rule, no need to check it here
        cnx = self.cnx
        pendingeids = cnx.transaction_data.get("pendingeids", ())
        if self.eid in pendingeids:
            return
        entity = cnx.entity_from_eid(self.eid)
        iworkflowable = entity.cw_adapt_to("IWorkflowable")
        # check custom workflow has not been rechanged to another one in the same
        # transaction
        mainwf = iworkflowable.main_workflow
        if mainwf.eid == self.wfeid:
            deststate = mainwf.initial
            if not deststate:
                msg = _("workflow has no initial state")
                raise validation_error(entity, {("custom_workflow", "subject"): msg})
            if mainwf.state_by_eid(iworkflowable.current_state.eid):
                # nothing to do
                return
            # if there are no history, simply go to new workflow's initial state
            if not iworkflowable.workflow_history:
                if iworkflowable.current_state.eid != deststate.eid:
                    _change_state(
                        cnx, entity.eid, iworkflowable.current_state.eid, deststate.eid
                    )
                    _FireAutotransitionOp(cnx, eid=entity.eid)
                return
            msg = cnx._('workflow changed to "%s"')
            msg %= cnx._(mainwf.name)
            cnx.transaction_data[(entity.eid, "customwf")] = self.wfeid
            iworkflowable.change_state(deststate, msg, "text/plain")


class _CheckTrExitPoint(hook.Operation):
    treid = None  # make pylint happy

    def precommit_event(self):
        tr = self.cnx.entity_from_eid(self.treid)
        outputs = set()
        for ep in tr.subworkflow_exit:
            if ep.subwf_state.eid in outputs:
                msg = _("can't have multiple exits on the same state")
                raise validation_error(
                    self.treid, {("subworkflow_exit", "subject"): msg}
                )
            outputs.add(ep.subwf_state.eid)


class _SubWorkflowExitOp(hook.Operation):
    foreid = trinfo = None  # make pylint happy

    def precommit_event(self):
        cnx = self.cnx
        forentity = cnx.entity_from_eid(self.foreid)
        iworkflowable = forentity.cw_adapt_to("IWorkflowable")
        trinfo = self.trinfo
        # we're in a subworkflow, check if we've reached an exit point
        wftr = iworkflowable.subworkflow_input_transition()
        if wftr is None:
            # inconsistency detected
            msg = _("state doesn't belong to entity's current workflow")
            raise validation_error(self.trinfo, {("to_state", "subject"): msg})
        tostate = wftr.get_exit_point(forentity, trinfo.cw_attr_cache["to_state"])
        if tostate is not None:
            # reached an exit point
            msg = _("exiting from subworkflow %s")
            msg %= cnx._(iworkflowable.current_workflow.name)
            cnx.transaction_data[(forentity.eid, "subwfentrytr")] = True
            iworkflowable.change_state(tostate, msg, "text/plain", tr=wftr)


# hooks ########################################################################


class WorkflowHook(hook.Hook):
    __abstract__ = True
    category = "metadata"


class SetInitialStateHook(WorkflowHook):
    __regid__ = "wfsetinitial"
    __select__ = WorkflowHook.__select__ & adaptable("IWorkflowable")
    events = ("after_add_entity",)

    def __call__(self):
        if self.entity.cw_etype == "CWUser" and self._cw.repo.config.creating:
            # do not set initial state for user during creating
            # the default workflow state for CWUser is added by
            # default in `cubicweb/misc/migration/postcreate.py`
            return
        _SetInitialStateOp(self._cw, eid=self.entity.eid)


class FireTransitionHook(WorkflowHook):
    """check the transition is allowed and add missing information into the
    TrInfo entity.

    Expect that:
    * wf_info_for inlined relation is set
    * by_transition or to_state (managers only) inlined relation is set

    Check for automatic transition to be fired at the end
    """

    __regid__ = "wffiretransition"
    __select__ = WorkflowHook.__select__ & is_instance("TrInfo")
    events = ("before_add_entity",)

    def __call__(self):
        cnx = self._cw
        entity = self.entity
        # first retreive entity to which the state change apply
        try:
            foreid = entity.cw_attr_cache["wf_info_for"]
        except KeyError:
            msg = _("mandatory relation")
            raise validation_error(entity, {("wf_info_for", "subject"): msg})
        forentity = cnx.entity_from_eid(foreid)
        # see comment in the TrInfo entity definition
        entity.cw_edited["tr_count"] = len(forentity.reverse_wf_info_for)
        iworkflowable = forentity.cw_adapt_to("IWorkflowable")
        # then check it has a workflow set, unless we're in the process of changing
        # entity's workflow
        if cnx.transaction_data.get((forentity.eid, "customwf")):
            wfeid = cnx.transaction_data[(forentity.eid, "customwf")]
            wf = cnx.entity_from_eid(wfeid)
        else:
            wf = iworkflowable.current_workflow
        if wf is None:
            msg = _("related entity has no workflow set")
            raise validation_error(entity, {None: msg})
        # then check it has a state set
        fromstate = iworkflowable.current_state
        if fromstate is None:
            msg = _("related entity has no state")
            raise validation_error(entity, {None: msg})
        # True if we are coming back from subworkflow
        swtr = cnx.transaction_data.pop((forentity.eid, "subwfentrytr"), None)
        cowpowers = cnx.user.is_in_group("managers") or not cnx.write_security
        # no investigate the requested state change...
        try:
            treid = entity.cw_attr_cache["by_transition"]
        except KeyError:
            # no transition set, check user is a manager and destination state
            # is specified (and valid)
            if not cowpowers:
                msg = _("mandatory relation")
                raise validation_error(entity, {("by_transition", "subject"): msg})
            deststateeid = entity.cw_attr_cache.get("to_state")
            if not deststateeid:
                msg = _("mandatory relation")
                raise validation_error(entity, {("by_transition", "subject"): msg})
            deststate = wf.state_by_eid(deststateeid)
            if deststate is None:
                msg = _("state doesn't belong to entity's workflow")
                raise validation_error(entity, {("to_state", "subject"): msg})
        else:
            # check transition is valid and allowed, unless we're coming back
            # from subworkflow
            tr = cnx.entity_from_eid(treid)
            if swtr is None:
                qname = ("by_transition", "subject")
                if tr is None:
                    msg = _("transition doesn't belong to entity's workflow")
                    raise validation_error(entity, {qname: msg})
                if not tr.has_input_state(fromstate):
                    msg = _("transition %(tr)s isn't allowed from %(st)s")
                    raise validation_error(
                        entity,
                        {qname: msg},
                        {"tr": tr.name, "st": fromstate.name},
                        ["tr", "st"],
                    )
                if not tr.may_be_fired(foreid):
                    msg = _("transition may not be fired")
                    raise validation_error(entity, {qname: msg})
            deststateeid = entity.cw_attr_cache.get("to_state")
            if deststateeid is not None:
                if not cowpowers and deststateeid != tr.destination(forentity).eid:
                    msg = _("transition isn't allowed")
                    raise validation_error(entity, {("by_transition", "subject"): msg})
                if swtr is None:
                    deststate = cnx.entity_from_eid(deststateeid)
                    if not cowpowers and deststate is None:
                        msg = _("state doesn't belong to entity's workflow")
                        raise validation_error(entity, {("to_state", "subject"): msg})
            else:
                deststateeid = tr.destination(forentity).eid
        # everything is ok, add missing information on the trinfo entity
        entity.cw_edited["from_state"] = fromstate.eid
        entity.cw_edited["to_state"] = deststateeid
        nocheck = cnx.transaction_data.setdefault("skip-security", set())
        nocheck.add((entity.eid, "from_state", fromstate.eid))
        nocheck.add((entity.eid, "to_state", deststateeid))
        _FireAutotransitionOp(cnx, eid=forentity.eid)


class FiredTransitionHook(WorkflowHook):
    """change related entity state and handle exit of subworkflow"""

    __regid__ = "wffiretransition"
    __select__ = WorkflowHook.__select__ & is_instance("TrInfo")
    events = ("after_add_entity",)

    def __call__(self):
        trinfo = self.entity
        rcache = trinfo.cw_attr_cache
        _change_state(
            self._cw, rcache["wf_info_for"], rcache["from_state"], rcache["to_state"]
        )
        forentity = self._cw.entity_from_eid(rcache["wf_info_for"])
        iworkflowable = forentity.cw_adapt_to("IWorkflowable")
        assert iworkflowable.current_state.eid == rcache["to_state"]
        if iworkflowable.main_workflow.eid != iworkflowable.current_workflow.eid:
            _SubWorkflowExitOp(self._cw, foreid=forentity.eid, trinfo=trinfo)


class CheckInStateChangeAllowed(WorkflowHook):
    """check state apply, in case of direct in_state change using unsafe execute"""

    __regid__ = "wfcheckinstate"
    __select__ = WorkflowHook.__select__ & hook.match_rtype("in_state")
    events = ("before_add_relation",)
    category = "integrity"

    def __call__(self):
        cnx = self._cw
        nocheck = cnx.transaction_data.get("skip-security", ())
        if (self.eidfrom, "in_state", self.eidto) in nocheck:
            # state changed through TrInfo insertion, so we already know it's ok
            return
        entity = cnx.entity_from_eid(self.eidfrom)
        iworkflowable = entity.cw_adapt_to("IWorkflowable")
        mainwf = iworkflowable.main_workflow
        if mainwf is None:
            msg = _("entity has no workflow set")
            raise validation_error(entity, {None: msg})
        for wf in mainwf.iter_workflows():
            if wf.state_by_eid(self.eidto):
                break
        else:
            msg = _(
                "state doesn't belong to entity's workflow. You may "
                "want to set a custom workflow for this entity first."
            )
            raise validation_error(self.eidfrom, {("in_state", "subject"): msg})
        if (
            iworkflowable.current_workflow
            and wf.eid != iworkflowable.current_workflow.eid
        ):
            msg = _("state doesn't belong to entity's current workflow")
            raise validation_error(self.eidfrom, {("in_state", "subject"): msg})


class SetModificationDateOnStateChange(WorkflowHook):
    """update entity's modification date after changing its state"""

    __regid__ = "wfsyncmdate"
    __select__ = WorkflowHook.__select__ & hook.match_rtype("in_state")
    events = ("after_add_relation",)

    def __call__(self):
        if self._cw.added_in_transaction(self.eidfrom):
            # new entity, not needed
            return
        entity = self._cw.entity_from_eid(self.eidfrom)
        try:
            entity.cw_set(modification_date=datetime.now(timezone.utc))
        except RepositoryError as ex:
            # usually occurs if entity is coming from a read-only source
            # (eg ldap user)
            self.warning("cant change modification date for %s: %s", entity, ex)


class CheckWorkflowTransitionExitPoint(WorkflowHook):
    """check that there is no multiple exits from the same state"""

    __regid__ = "wfcheckwftrexit"
    __select__ = WorkflowHook.__select__ & hook.match_rtype("subworkflow_exit")
    events = ("after_add_relation",)

    def __call__(self):
        _CheckTrExitPoint(self._cw, treid=self.eidfrom)


class SetCustomWorkflow(WorkflowHook):
    __regid__ = "wfsetcustom"
    __select__ = WorkflowHook.__select__ & hook.match_rtype("custom_workflow")
    events = ("after_add_relation",)

    def __call__(self):
        _WorkflowChangedOp(self._cw, eid=self.eidfrom, wfeid=self.eidto)


class DelCustomWorkflow(SetCustomWorkflow):
    __regid__ = "wfdelcustom"
    events = ("after_delete_relation",)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        typewf = entity.cw_adapt_to("IWorkflowable").cwetype_workflow()
        if typewf is not None:
            _WorkflowChangedOp(self._cw, eid=self.eidfrom, wfeid=typewf.eid)



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.mail"""

import os
import re
from socket import gethostname
from unittest import TestCase

from logilab.common.umessage import message_from_string

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.mail import format_mail, construct_message_id, parse_message_id


def getlogin():
    """avoid using os.getlogin() because of strange tty / stdin problems
    (man 3 getlogin)
    Another solution would be to use $LOGNAME, $USER or $USERNAME
    """
    import pwd

    return pwd.getpwuid(os.getuid())[0]


class EmailTC(CubicWebTC):
    def test_format_mail(self):
        self.set_option("sender-addr", "bim@boum.fr")
        self.set_option("sender-name", "BimBam")

        mail = format_mail(
            {"name": "oim", "email": "oim@logilab.fr"},
            ["test@logilab.fr"],
            "un petit cöucou",
            "bïjour",
            config=self.config,
        )
        result = mail.as_string()
        result = re.sub("^Date: .*$", "Date: now", result, flags=re.MULTILINE)
        self.assertMultiLineEqual(
            result,
            """\
MIME-Version: 1.0
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: base64
Subject: =?utf-8?q?b=C3=AFjour?=
From: =?utf-8?q?oim?= <bim@boum.fr>
X-CW: data
To: test@logilab.fr
Date: now

dW4gcGV0aXQgY8O2dWNvdQ==
""",
        )
        msg = message_from_string(mail.as_string())
        self.assertEqual(msg.get("subject"), "bïjour")
        self.assertEqual(msg.get("from"), "oim <bim@boum.fr>")
        self.assertEqual(msg.get("to"), "test@logilab.fr")
        self.assertIsNone(msg.get("reply-to"))
        self.assertEqual(msg.get_payload(decode=True), "un petit cöucou")

    def test_format_mail_euro(self):
        mail = format_mail(
            {"name": "oîm", "email": "oim@logilab.fr"},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
        )
        result = mail.as_string()
        result = re.sub("^Date: .*$", "Date: now", result, flags=re.MULTILINE)
        self.assertMultiLineEqual(
            result,
            """\
MIME-Version: 1.0
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: base64
Subject: =?utf-8?b?YsOvam91ciDigqw=?=
From: =?utf-8?q?o=C3=AEm?= <oim@logilab.fr>
To: test@logilab.fr
Date: now

dW4gcGV0aXQgY8O2dWNvdSDigqw=
""",
        )
        msg = message_from_string(mail.as_string())
        self.assertEqual(msg.get("subject"), "bïjour €")
        self.assertEqual(msg.get("from"), "oîm <oim@logilab.fr>")
        self.assertEqual(msg.get("to"), "test@logilab.fr")
        self.assertEqual(msg.get_payload(decode=True), "un petit cöucou €")

    def test_format_mail_from_reply_to_is_none(self):
        # no sender-name, sender-addr in the configuration
        self.set_option("sender-name", "")
        self.set_option("sender-addr", "")
        msg = format_mail(
            {"name": "", "email": ""},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
            config=self.config,
        )
        self.assertEqual(msg.get("from"), "")
        self.assertIsNone(msg.get("reply-to"))

        msg = format_mail(
            {"name": "tutu", "email": "tutu@logilab.fr"},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
            config=self.config,
        )
        msg = message_from_string(msg.as_string())
        self.assertEqual(msg.get("from"), "tutu <tutu@logilab.fr>")
        self.assertIsNone(msg.get("reply-to"))

        msg = format_mail(
            {"name": "tutu", "email": "tutu@logilab.fr"},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
        )
        msg = message_from_string(msg.as_string())
        self.assertEqual(msg.get("from"), "tutu <tutu@logilab.fr>")
        self.assertIsNone(msg.get("reply-to"))

        # set sender name and address as expected
        self.set_option("sender-name", "cubicweb-test")
        self.set_option("sender-addr", "cubicweb-test@logilab.fr")
        # anonymous notification: no name and no email specified
        msg = format_mail(
            {"name": "", "email": ""},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
            config=self.config,
        )
        msg = message_from_string(msg.as_string())
        self.assertEqual(msg.get("from"), "cubicweb-test <cubicweb-test@logilab.fr>")
        self.assertIsNone(msg.get("reply-to"))

        # anonymous notification: only email specified
        msg = format_mail(
            {"email": "tutu@logilab.fr"},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
            config=self.config,
        )
        msg = message_from_string(msg.as_string())
        self.assertEqual(msg.get("from"), "cubicweb-test <cubicweb-test@logilab.fr>")
        self.assertIsNone(msg.get("reply-to"))

        # anonymous notification: only name specified
        msg = format_mail(
            {"name": "tutu"},
            ["test@logilab.fr"],
            "un petit cöucou €",
            "bïjour €",
            config=self.config,
        )
        msg = message_from_string(msg.as_string())
        self.assertEqual(msg.get("from"), "tutu <cubicweb-test@logilab.fr>")
        self.assertIsNone(msg.get("reply-to"))


class MessageIdTC(TestCase):
    def test_base(self):
        msgid1 = construct_message_id("testapp", 21)
        msgid2 = construct_message_id("testapp", 21)
        self.assertNotEqual(msgid1, msgid2)
        self.assertNotIn("&", msgid1)
        self.assertNotIn("=", msgid1)
        self.assertNotIn("/", msgid1)
        self.assertNotIn("+", msgid1)
        values = parse_message_id(msgid1, "testapp")
        self.assertTrue(values)
        # parse_message_id should work with or without surrounding <>
        self.assertEqual(values, parse_message_id(msgid1[1:-1], "testapp"))
        self.assertEqual(values["eid"], "21")
        self.assertIn("timestamp", values)
        self.assertEqual(parse_message_id(msgid1[1:-1], "anotherapp"), None)

    def test_notimestamp(self):
        msgid1 = construct_message_id("testapp", 21, False)
        construct_message_id("testapp", 21, False)
        values = parse_message_id(msgid1, "testapp")
        self.assertEqual(values, {"eid": "21"})

    def test_parse_message_doesnt_raise(self):
        self.assertEqual(parse_message_id("oijioj@bla.bla", "tesapp"), None)
        self.assertEqual(parse_message_id("oijioj@bla", "tesapp"), None)
        self.assertEqual(parse_message_id("oijioj", "tesapp"), None)

    def test_nonregr_empty_message_id(self):
        for eid in (1, 12, 123, 1234):
            msgid1 = construct_message_id("testapp", eid, 12)
            self.assertNotEqual(msgid1, f"<@testapp.{gethostname()}>")


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""utilities for instances migration"""

import itertools
import logging
import os
import string
import sys
import tempfile
import traceback
from itertools import chain
from os.path import exists, join, basename, splitext

from logilab.common import IGNORED_EXTENSIONS
from logilab.common.changelog import Version
from logilab.common.configuration import REQUIRED, read_old_config
from logilab.common.decorators import cached
from logilab.common.shellutils import ASK

from cubicweb import ConfigurationError, ExecutionError, utils, set_log_methods
from cubicweb.cwconfig import CubicWebConfiguration as cwcfg, CONFIG_HELP_MESSAGE
from cubicweb.toolsutils import show_diffs


def filter_scripts(config, directory, fromversion, toversion, quiet=True):
    """return a list of paths of migration files to consider to upgrade
    from a version to a greater one
    """
    from logilab.common.changelog import Version  # doesn't work with appengine

    assert fromversion
    assert toversion
    assert isinstance(fromversion, tuple), fromversion.__class__
    assert isinstance(toversion, tuple), toversion.__class__
    assert fromversion <= toversion, (fromversion, toversion)
    if not exists(directory):
        if not quiet:
            print(directory, "doesn't exists, no migration path")
        return []
    if fromversion == toversion:
        return []
    result = []
    for fname in os.listdir(directory):
        if fname.endswith(IGNORED_EXTENSIONS):
            continue
        fpath = join(directory, fname)
        try:
            tver, mode = fname.split("_", 1)
        except ValueError:
            continue
        mode = mode.split(".", 1)[0]
        if not config.accept_mode(mode):
            continue
        try:
            tver = Version(tver)
        except ValueError:
            continue
        if tver <= fromversion:
            continue
        if tver > toversion:
            continue
        result.append((tver, fpath))
    # be sure scripts are executed in order
    return sorted(result)


def execscript_confirm(scriptpath):
    """asks for confirmation before executing a script and provides the
    ability to show the script's content
    """
    while True:
        answer = ASK.ask(f"Execute {scriptpath!r} ?", ("Y", "n", "show", "abort"), "Y")
        if answer == "abort":
            raise SystemExit(1)
        elif answer == "n":
            return False
        elif answer == "show":
            stream = open(scriptpath)
            scriptcontent = stream.read()
            stream.close()
            print()
            print(scriptcontent)
            print()
        else:
            return True


def yes(*args, **kwargs):
    return True


class MigrationHelper:
    """class holding CubicWeb Migration Actions used by migration scripts"""

    def __init__(self, config, interactive=True, verbosity=1):
        self.config = config
        if config:
            # no config on shell to a remote instance
            self.config.init_log(logthreshold=logging.ERROR)
        # 0: no confirmation, 1: only main commands confirmed, 2 ask for everything
        self.verbosity = verbosity
        self.need_wrap = True
        if not interactive or not verbosity:
            self.confirm = yes
            self.execscript_confirm = yes
        else:
            self.execscript_confirm = execscript_confirm
        self._option_changes = []
        self.__context = {
            "confirm": self.confirm,
            "config": self.config,
            "interactive_mode": interactive,
        }
        self._context_stack = []

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            cmd = f"cmd_{name}"
            # search self.__class__ to avoid infinite recursion
            if hasattr(self.__class__, cmd):
                meth = getattr(self, cmd)
                try:
                    return lambda *args, **kwargs: self.interact(
                        args, kwargs, meth=meth
                    )
                except Exception:
                    _, ex, traceback_ = sys.exc_info()
                    traceback.print_exc()
                    if self.confirm("abort?", pdb=True, traceback=traceback_):
                        raise
            raise
        raise AttributeError(name)

    def migrate(self, vcconf, toupgrade, options):
        """upgrade the given set of cubes

        `cubes` is an ordered list of 3-uple:
        (cube, fromversion, toversion)
        """
        if options.fs_only:
            # monkey path configuration.accept_mode so database mode (e.g. Any)
            # won't be accepted
            orig_accept_mode = self.config.accept_mode

            def accept_mode(mode):
                if mode == "Any":
                    return False
                return orig_accept_mode(mode)

            self.config.accept_mode = accept_mode
        # may be an iterator
        toupgrade = tuple(toupgrade)
        vmap = {cube: (fromver, tover) for cube, fromver, tover in toupgrade}
        ctx = self.__context
        ctx["versions_map"] = vmap
        if self.config.accept_mode("Any") and "cubicweb" in vmap:
            migrdir = self.config.migration_scripts_dir()
            self.cmd_process_script(join(migrdir, "bootstrapmigration_repository.py"))
        for cube, fromversion, toversion in toupgrade:
            if cube == "cubicweb":
                migrdir = self.config.migration_scripts_dir()
            else:
                migrdir = self.config.cube_migration_scripts_dir(cube)
            scripts = filter_scripts(self.config, migrdir, fromversion, toversion)
            if scripts:
                prevversion = None
                for version, script in scripts:
                    # take care to X.Y.Z_Any.py / X.Y.Z_common.py: we've to call
                    # cube_upgraded once all script of X.Y.Z have been executed
                    if prevversion is not None and version != prevversion:
                        self.cube_upgraded(cube, prevversion)
                    prevversion = version
                    self.cmd_process_script(script)
                self.cube_upgraded(cube, toversion)
            else:
                self.cube_upgraded(cube, toversion)

    def cube_upgraded(self, cube, version):
        pass

    def shutdown(self):
        pass

    def interact(self, args, kwargs, meth):
        """execute the given method according to user's confirmation"""
        msg = "Execute command: {}({}) ?".format(
            meth.__name__[4:],
            ", ".join(
                [repr(arg) for arg in args] + [f"{n}={v!r}" for n, v in kwargs.items()]
            ),
        )
        if "ask_confirm" in kwargs:
            ask_confirm = kwargs.pop("ask_confirm")
        else:
            ask_confirm = True
        if not ask_confirm or self.confirm(msg):
            return meth(*args, **kwargs)

    def confirm(
        self,
        question,  # pylint: disable=E0202
        shell=True,
        abort=True,
        retry=False,
        pdb=False,
        default="y",
        traceback=None,
    ):
        """ask for confirmation and return true on positive answer

        if `retry` is true the r[etry] answer may return 2
        """
        possibleanswers = ["y", "n"]
        if abort:
            possibleanswers.append("abort")
        if pdb:
            possibleanswers.append("pdb")
        if shell:
            possibleanswers.append("shell")
        if retry:
            possibleanswers.append("retry")
        try:
            answer = ASK.ask(question, possibleanswers, default)
        except (EOFError, KeyboardInterrupt):
            answer = "abort"
        if answer == "n":
            return False
        if answer == "retry":
            return 2
        if answer == "abort":
            raise SystemExit(1)
        if answer == "shell":
            self.interactive_shell()
            return self.confirm(question, shell, abort, retry, pdb, default, traceback)
        if answer == "pdb":
            pdb = utils.get_pdb()
            if traceback:
                pdb.post_mortem(traceback)
            else:
                pdb.set_trace()
            return self.confirm(question, shell, abort, retry, pdb, default, traceback)
        return True

    def interactive_shell(self):
        self.confirm = yes
        self.need_wrap = False

        # avoid '_' to be added to builtins by sys.display_hook
        def do_not_add___to_builtins(obj):
            if obj is not None:
                print(repr(obj))

        sys.displayhook = do_not_add___to_builtins
        local_ctx = self._create_context()
        try:
            import readline
            from cubicweb.toolsutils import CWShellCompleter
        except ImportError:
            # readline not available
            pass
        else:
            rql_completer = CWShellCompleter(local_ctx)
            readline.set_completer(rql_completer.complete)
            readline.parse_and_bind("tab: complete")
            home_key = "HOME"
            histfile = os.path.join(os.environ[home_key], ".cwshell_history")
            try:
                readline.read_history_file(histfile)
            except OSError:
                pass
        from code import interact

        banner = """entering the migration python shell
just type migration commands or arbitrary python code and type ENTER to execute it
type "exit" or Ctrl-D to quit the shell and resume operation"""

        # use ipython if available
        try:
            from IPython import start_ipython

            print(banner)
            start_ipython(argv=[], user_ns=local_ctx)
        except ImportError:
            interact(banner, local=local_ctx)

            try:
                readline.write_history_file(histfile)
            except OSError:
                pass

        # delete instance's confirm attribute to avoid questions
        del self.confirm
        self.need_wrap = True

    @cached
    def _create_context(self):
        """return a dictionary to use as migration script execution context"""
        context = self.__context
        for attr in dir(self):
            if attr.startswith("cmd_"):
                if self.need_wrap:
                    context[attr[4:]] = getattr(self, attr[4:])
                else:
                    context[attr[4:]] = getattr(self, attr)
        return context

    def update_context(self, key, value):
        for context in self._context_stack:
            context[key] = value
        self.__context[key] = value

    def cmd_process_script(self, migrscript, funcname=None, *args, **kwargs):
        """execute a migration script in interactive mode

        Display the migration script path, ask for confirmation and execute it
        if confirmed

        Allowed input file formats for migration scripts:
        - `python` (.py)
        - `sql` (.sql)
        - `doctest` (.txt or .rst)

        .. warning:: sql migration scripts are not available in web-only instance

        You can pass script parameters with using double dash (--) in the
        command line

        Context environment can have these variables defined:
        - __name__ : will be determine by funcname parameter
        - __file__ : is the name of the script if it exists
        - __args__ : script arguments coming from command-line

        :param migrscript: name of the script
        :param funcname: defines __name__ inside the shell (or use __main__)
        :params args: optional arguments for funcname
        :keyword scriptargs: optional arguments of the script
        """
        ftypes = {"python": (".py",), "doctest": (".txt", ".rst"), "sql": (".sql",)}
        # sql migration scripts are not available in web-only instance
        if not hasattr(self, "session"):
            ftypes.pop("sql")
        migrscript = os.path.normpath(migrscript)
        for script_mode, ftype in ftypes.items():
            if migrscript.endswith(ftype):
                break
        else:
            ftypes = ", ".join(chain(*ftypes.values()))
            msg = "ignoring %s, not a valid script extension (%s)"
            raise ExecutionError(msg % (migrscript, ftypes))
        if not self.execscript_confirm(migrscript):
            return
        scriptlocals = self._create_context().copy()
        scriptlocals.update(
            {"__file__": migrscript, "__args__": kwargs.pop("scriptargs", [])}
        )
        self._context_stack.append(scriptlocals)
        if script_mode == "python":
            if funcname is None:
                pyname = "__main__"
            else:
                pyname = splitext(basename(migrscript))[0]
            scriptlocals["__name__"] = pyname
            with open(migrscript, "rb") as fobj:
                fcontent = fobj.read()
            code = compile(fcontent, migrscript, "exec")
            exec(code, scriptlocals)
            if funcname is not None:
                try:
                    func = scriptlocals[funcname]
                    self.info("found %s in locals", funcname)
                    assert callable(func), f"{func} ({funcname}) is not callable"
                except KeyError:
                    self.critical("no %s in script %s", funcname, migrscript)
                    return None
                return func(*args, **kwargs)
        elif script_mode == "sql":
            from cubicweb.server.sqlutils import sqlexec

            sqlexec(open(migrscript).read(), self.session.system_sql)
            self.commit()
        else:  # script_mode == 'doctest'
            import doctest

            return doctest.testfile(
                migrscript,
                module_relative=False,
                optionflags=doctest.ELLIPSIS,
                # verbose mode when user input is expected
                verbose=self.verbosity == 2,
                report=True,
                encoding="utf-8",
                globs=scriptlocals,
            )
        self._context_stack.pop()

    def cmd_option_renamed(self, oldname, newname):
        """a configuration option has been renamed"""
        self._option_changes.append(("renamed", oldname, newname))

    def cmd_option_group_changed(self, option, oldgroup, newgroup):
        """a configuration option has been moved in another group"""
        self._option_changes.append(("moved", option, oldgroup, newgroup))

    def cmd_option_added(self, optname):
        """a configuration option has been added"""
        self._option_changes.append(("added", optname))

    def cmd_option_removed(self, optname):
        """a configuration option has been removed"""
        # can safely be ignored
        # self._option_changes.append(('removed', optname))

    def cmd_option_type_changed(self, optname, oldtype, newvalue):
        """a configuration option's type has changed"""
        self._option_changes.append(("typechanged", optname, oldtype, newvalue))

    def cmd_add_cubes(self, cubes):
        """modify the list of used cubes in the in-memory config
        returns newly inserted cubes, including dependencies
        """
        if isinstance(cubes, str):
            cubes = (cubes,)
        origcubes = self.config.cubes()
        newcubes = [p for p in self.config.expand_cubes(cubes) if p not in origcubes]
        if newcubes:
            self.config.add_cubes(newcubes)
        return newcubes

    def cmd_drop_cube(self, cube, removedeps=False, force_drop=False):
        if removedeps:
            toremove = self.config.expand_cubes([cube])
        else:
            toremove = (cube,)
        origcubes = self.config._cubes
        basecubes = [c for c in origcubes if c not in toremove]
        # don't fake-add any new ones, or we won't be able to really-add them later
        new_config_cubes = []
        for c in self.config.expand_cubes(basecubes):
            if c in origcubes:
                if c in toremove and force_drop:
                    continue
                new_config_cubes.append(c)
        self.config._cubes = tuple(new_config_cubes)
        removed = [p for p in origcubes if p not in self.config._cubes]
        if cube not in removed and cube in origcubes:
            raise ConfigurationError(
                f"can't remove cube {cube}, used as a dependency."
                f"Consider using 'force_drop=True' if you want to drop it anyway"
            )
        return removed

    def rewrite_configuration(self):
        configfile = self.config.main_config_file()
        if self._option_changes:
            read_old_config(self.config, self._option_changes, configfile)
        fd, newconfig = tempfile.mkstemp()
        for optdescr in self._option_changes:
            if optdescr[0] == "added":
                optdict = self.config.get_option_def(optdescr[1])
                if optdict.get("default") is REQUIRED:
                    self.config.input_option(optdescr[1], optdict)
        self.config.generate_config(
            open(newconfig, "w"), header_message=CONFIG_HELP_MESSAGE
        )
        show_diffs(configfile, newconfig, askconfirm=self.confirm is not yes)
        os.close(fd)
        if exists(newconfig):
            os.unlink(newconfig)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


set_log_methods(MigrationHelper, logging.getLogger("cubicweb.migration"))


def version_strictly_lower(a, b):
    if a is None:
        return True
    if b is None:
        return False
    if a:
        a = Version(a)
    if b:
        b = Version(b)
    return a < b


def max_version(a, b):
    return str(max(Version(a), Version(b)))


def split_constraint(constraint):
    oper = itertools.takewhile(lambda x: x in "<>=", constraint)
    version = itertools.dropwhile(lambda x: x not in string.digits + ".", constraint)
    return "".join(oper), "".join(version)


def parse_constraints(stream):
    constraints = []
    for block in stream.split(","):
        block = block.strip()
        constraints.append(split_constraint(block))
    return constraints


class ConfigurationProblem:
    """Each cube has its own list of dependencies on other cubes/versions.

    The ConfigurationProblem is used to record the loaded cubes, then to detect
    inconsistencies in their dependencies.

    See configuration management on wikipedia for litterature.
    """

    def __init__(self, config):
        self.config = config
        self.cubes = {"cubicweb": cwcfg.cubicweb_version()}

    def add_cube(self, name, version):
        self.cubes[name] = version

    def solve(self):
        self.warnings = []
        self.errors = []
        self.dependencies = {}
        self.reverse_dependencies = {}
        self.constraints = {}
        # read dependencies
        for cube in self.cubes:
            if cube == "cubicweb":
                continue
            self.dependencies[cube] = dict(self.config.cube_dependencies(cube))
            self.dependencies[cube]["cubicweb"] = (
                self.config.cube_depends_cubicweb_version(cube)
            )
        # compute reverse dependencies
        for cube, dependencies in self.dependencies.items():
            for name, constraints in dependencies.items():
                self.reverse_dependencies.setdefault(name, set())
                if constraints:
                    try:
                        for oper, version in parse_constraints(constraints):
                            self.reverse_dependencies[name].add((oper, version, cube))
                    except Exception:
                        self.warnings.append(
                            "cube %s depends on %s but constraint badly "
                            "formatted: %s" % (cube, name, constraints)
                        )
                else:
                    self.reverse_dependencies[name].add((None, None, cube))
        # check consistency
        for cube, versions in sorted(self.reverse_dependencies.items()):
            min_version, min_source = None, None
            max_version, max_source = None, None
            for oper, version, source in versions:
                if oper == ">=":
                    if version_strictly_lower(min_version, version):
                        min_version = version
                        min_source = source
                elif oper == "<":
                    if version_strictly_lower(version, max_version):
                        max_version = version
                        max_source = source
                elif oper is None:
                    pass
                else:
                    print(
                        "unable to handle %s in %s, set to `%s %s` "
                        "but currently up to `%s %s`"
                        % (cube, source, oper, version, min_version, max_version)
                    )
            if cube not in self.cubes:
                self.errors.append(("add", cube, min_version, min_source))
                continue
            installed_version = self.cubes[cube]
            if min_version and version_strictly_lower(installed_version, min_version):
                self.errors.append(("update", cube, min_version, min_source))
            if max_version and version_strictly_lower(max_version, installed_version):
                self.errors.append(("update", cube, max_version, max_source))
